from io import BytesIO
import graphviz
import matplotlib

matplotlib.use('Agg')
import os
import shutil
import tempfile

from decimal import Decimal
import matplotlib.ticker as ticker
from matplotlib.colors import ListedColormap
import matplotlib.patches as patches
import matplotlib.pyplot as plt
from matplotlib_venn import venn2, venn3, venn2_circles, venn3_circles

import numpy as np
import pandas as pd
from pathlib import Path
import seaborn as sns
from scipy import interp
from skimage.transform import resize
from skimage import color as skcolor
from sklearn.metrics import auc


from ..model.BreathCore import PeakAlignmentMethod, MccImsAnalysis, MccImsMeasurement
from ..model.BreathCore import deprecated


# Collection of plotting functions
class HeatmapPlot(object):
    """
    Collection of Heatmap plots
    """


    @staticmethod
    def prepare_fast_heatmap_plot(df, cmap_dict):
        fig = plt.figure()
        ax = fig.add_subplot(111)
        # faster conversion
        image = skcolor.rgb2gray(df.T.values)
        # image.shape
        x_size, y_size = 2500, 2500
        image_resized = resize(image, (x_size, y_size), mode='reflect')
        plt.imshow(image_resized, **cmap_dict)
        ax.invert_yaxis()
        plt.colorbar()

        # to plot we need to use the indices of the intensity matrix - not the actual values
        irm_vals = np.asarray(df.index.values, dtype=float)
        rt_vals = np.asarray(df.T.index.values, dtype=float)
        # need to re-adjust the stepwidth - as we resized the image
        ClusterPlot.setup_axes_for_fast_intensity_matrix(ax, irm_vals=irm_vals, rt_vals=rt_vals, x_size=x_size,
                                                         y_size=y_size)
        fig.set_figsize = (9, 4)
        return fig, ax


    @staticmethod
    def plot_heatmap_helper_fast(mcc_ims_measurement, plot_parameters, title="", title_prefix="Measurement ",
                                 plot_dir_suffix="heatmaps", plot_type="intentsity_plot"):
        intensity_matrix = mcc_ims_measurement.df
        measurement_name = mcc_ims_measurement.filename
        # fig = plt.figure()
        # ax = fig.add_subplot(111)
        # # faster conversion
        # image = skcolor.rgb2gray(intensity_matrix.T.values)
        # # image.shape
        # x_size, y_size = 2500, 2500
        # image_resized = resize(image, (x_size, y_size), mode='reflect')
        # plt.imshow(image_resized, cmap=plot_parameters['colormap'], vmin=0, vmax=1)
        # ax.invert_yaxis()
        # plt.colorbar()
        #
        # # to plot we need to use the indices of the intensity matrix - not the actual values
        # irm_vals = np.asarray(intensity_matrix.index.values, dtype=float)
        # rt_vals = np.asarray(intensity_matrix.T.index.values, dtype=float)
        # # need to re-adjust the stepwidth - as we resized the image
        # ClusterPlot.setup_axes_for_fast_intensity_matrix(ax, irm_vals=irm_vals, rt_vals=rt_vals, x_size=x_size, y_size=y_size)
        # fig.set_figsize = (9, 4)
        #
        fig, ax = HeatmapPlot.prepare_fast_heatmap_plot(intensity_matrix, {"cmap":plot_parameters['colormap'], "vmin":0, "vmax":1})
        if not title:
            title_string = '{}{}'.format(title_prefix, measurement_name)
        else:
            title_string = title

        plt.yticks(rotation=0)
        ax.set_title(title_string)

        figure_dir = '{}{}/'.format(plot_parameters['plot_dir'], plot_dir_suffix)

        plot_prefix = ''
        if 'plot_prefix' in plot_parameters:
            plot_prefix = plot_parameters['plot_prefix']
            plot_prefix = "fast_{}".format(plot_prefix)
            # import pdb;pdb.set_trace()
        dataset_name = plot_prefix

        figure_name = '{}_{}_{}.png'.format(dataset_name, plot_type, measurement_name[:-4])

        if not plot_parameters.get('use_buffer', False):
            Path(figure_dir).mkdir(parents=True, exist_ok=True)
            print(f"Saving figure to {figure_dir}{figure_name}")
            fig.savefig("{}{}".format(figure_dir, figure_name), dpi=200, bbox_inches='tight', format="png",
                        compress_level=1)
        return_figure = save_plot_to_buffer(plot_parameters, fig)
        plt.close()
        return return_figure


    @staticmethod
    def plot_classwise_heatmaps(mccImsAnalysis, plot_parameters):
        """
        Group Measurements by their class label and create an averaged heatmap
        :type mccImsAnalysis: MccImsAnalysis or list(model.BreathCore.MccImsMeasurement) with assigned class_labels
        :param mccImsAnalysis: MccImsAnalysis holding the class labels and measurements
        :param plot_parameters: dictionary holding plotting parameters
        :return: [tuple(Buffer,....) / None], saves consensus heatmaps
        """
        plots_to_return = []
        if isinstance(mccImsAnalysis, MccImsAnalysis):
            class_labels = np.unique([m.class_label for m in mccImsAnalysis.measurements])
            class_label_measurement_dict = {cl: [] for cl in class_labels}
            for m in mccImsAnalysis.measurements:
                class_label_measurement_dict[m.class_label].append(m)

        elif isinstance(mccImsAnalysis, list):
            class_labels = np.unique([m.class_label for m in mccImsAnalysis])
            class_label_measurement_dict = {cl: [] for cl in class_labels}
            for m in mccImsAnalysis:
                class_label_measurement_dict[m.class_label].append(m)
        else:
            raise ValueError("Expected MccImsAnalysis or list of measurements, not {}".format(type(mccImsAnalysis)))

        print(class_label_measurement_dict)
        for k,v in class_label_measurement_dict.items():
            plots_to_return.append(HeatmapPlot._plot_classwise_heatmap_helper(v, k, plot_parameters))
        return plots_to_return
    
    @staticmethod
    def compute_averaged_df(dfs):
        averaged_df = pd.DataFrame()
        # not very simple, need to do index things - need to find fewest rows and fewest columns, then crop all of them to same dimensions
        # then drop indices and add to each other
        # then assign original index
        # is_measurement_list = isinstance(dfs[0], MccImsMeasurement)
        if isinstance(dfs[0], pd.DataFrame):
            # real_indices = [df.index for df in dfs]
            lowest_x = max(0.0, min([df.index.min() for df in dfs]))
            lowest_highest_x = min([df.index.max() for df in dfs])
            lowest_y = max(0.0, min([df.T.index.min() for df in dfs]))
            lowest_highest_y = max(0.0, min([df.T.index.max() for df in dfs]))
        else:
            raise ValueError("Expected list of MccImsMeasurements or pandas.DataFrames, not [{}]".format(
                type(dfs[0])))

        averaged_index, transposed_average_index = None, None
        for i, df in enumerate(dfs):
            # if is_measurement_list:
            selected_df = df[np.logical_and(df.index >= lowest_x, df.index <= lowest_highest_x)]
            selected_df_transp = selected_df.T
            selected_df_transp = selected_df_transp[
                np.logical_and(selected_df_transp.index >= lowest_y, selected_df_transp.index <= lowest_highest_y)]

            # save first index for later reassignment
            if not i:
                averaged_index = selected_df.index.copy()
                transposed_average_index = selected_df_transp.index.copy()

            selected_df_transp.reset_index(inplace=True, drop=True)
            selected_df = selected_df_transp.T

            # pandas really tries to make adding non-complying shapes / indices hard -
            # so we need to adjust the dimensions of the df even more -> pad with 0.0 columns / rows to the end
            # or delete last column/row to fit shape
            # print(selected_df.shape)
            selected_df.reset_index(inplace=True, drop=True)
            # print(selected_df.index)
            # print(selected_df.T.index)
            if i:
                if average_shape[0] > selected_df.shape[0]:
                    selected_df[len(selected_df)] = 0.0
                elif average_shape[0] < selected_df.shape[0]:
                    appropriate_len = average_shape[0] - 1
                    selected_df = selected_df[selected_df.index <= appropriate_len]

                if average_shape[1] > selected_df.shape[1]:
                    selected_df.T[average_shape[1]] = 0.0
                elif average_shape[1] < selected_df.shape[1]:
                    appropriate_len = average_shape[1] - 1
                    selected_df = selected_df.T[selected_df.T.index <= appropriate_len]
                    selected_df = selected_df.T

                averaged_df += selected_df
            else:
                averaged_df = selected_df
            average_shape = averaged_df.shape

        averaged_df /= len(dfs)

        # but we need an index for our heatmap plot --> reassign index with index of first element
        averaged_df.set_index(averaged_index, drop=True, inplace=True)
        averaged_df_t = averaged_df.T
        averaged_df_t.set_index(transposed_average_index, drop=True, inplace=True)
        return averaged_df_t.T

    @staticmethod
    def _plot_classwise_heatmap_helper(mccImsMeasurements, class_label, plot_parameters):
        """
        Compute average for grouped Measurements and return buffer or None
        :type mccImsMeasurements: [MccImsMeasurements,...]
        :param mccImsMeasurements: mccImsMeasurements each with the same class label
        :param plot_parameters: dictionary holding plotting parameters
        :return: tuple(Buffer,....) / None, saves averaged heatmaps
        """
        
        averaged_df = HeatmapPlot.compute_averaged_df([m.df for m in mccImsMeasurements])
        
        dummy_measurement = MccImsMeasurement("{} ".format(class_label), init_df=averaged_df, class_label=class_label)
        descriptor = "Averaged_class {}".format(class_label)
        return (class_label, HeatmapPlot.plot_heatmap_helper_fast(mcc_ims_measurement=dummy_measurement, plot_parameters=plot_parameters, title=descriptor, plot_type="classwise_heatmap",))

    @staticmethod
    def plot_feature_matrix(feature_matrix, class_label_dict, plot_parameters, outname=""):
        """
        TODO needs annotation
        Visualize the feature matrix by class using the normalized intensity of each feature
        :param feature_matrix:
        :param class_label_dict:
        :param plot_parameters:
        :param outname:
        :return:
        """
        from collections import defaultdict
        fig = plt.figure()
        # fig = ax.get_figure()
        # iterate over all class labels - filter rows by correct label and create subplot
        class_to_row_map = defaultdict(list)
        for rn, cl in class_label_dict.items():
            class_to_row_map[cl].append(rn)

        num_loops = len(class_to_row_map.keys())
        for i, (cl, row_selection) in enumerate(class_to_row_map.items()):
            # create mask for selection of correct rows
            selection_mask = [i in row_selection for i in feature_matrix.index]
            relevant_rows = feature_matrix.loc[selection_mask]

            ax = fig.add_subplot(2,1,i+1)
            # force all x labels...
            sns.heatmap(relevant_rows, ax=ax, xticklabels=True)
            ax.set_ylabel(cl)
            # remove x-axis label for all axes except the last one
            if i < num_loops-1: # enumerate starts at 0 and will end at len(-1)
                ax.get_xaxis().set_ticks([])
            # else:
            #     ax.get_xaxis().set_ticks(relevant_rows.columns.values)

        # fig, ax = HeatmapPlot.prepare_fast_heatmap_plot(feature_matrix,
        #                                                 {"cmap": plot_parameters['colormap'], "vmin": 0, "vmax": 1})

        # plt.yticks(rotation=0)
        if not outname:
            plot_suffix = "fm_plot",
            figure_dir = f"{plot_parameters['plot_dir']}{plot_suffix}"
            figure_name = f'{plot_suffix}.png'
        else:
            figure_dir = Path(outname).parent
            figure_name= Path(outname).stem + Path(outname).suffix

        # make subplot for all rows of certain class
        if not plot_parameters.get('use_buffer', False):
            Path(figure_dir).mkdir(parents=True, exist_ok=True)
            print(f"Saving figure to {figure_dir}/{figure_name}")
            fig.savefig(f"{figure_dir}/{figure_name}", dpi=300, bbox_inches='tight', format="png",
                        compress_level=1)
        return_figure = save_plot_to_buffer(plot_parameters, fig)
        plt.close()
        return return_figure


    # IntensityMatrix = _plot_heatmap_from_mcc_ims_measurement
    # BasicHeatmap = plot_heatmap_helper
    FastIntensityMatrix = plot_heatmap_helper_fast
    ClasswiseHeatmaps = plot_classwise_heatmaps
    FeatureMatrixPlot = plot_feature_matrix
    # MultipleIntensityMatrices = plot_intensity_matrices


class ClusterPlot(object):
    """
    Collection of Cluster plots
    """
    @staticmethod
    def _map_clustering_parameters_to_string(param_dict, method_name):
        """
        Takes clustering parameters and constructs them to a String
        :param param_dict:
        :param method_name:
        :return:
        """
        method_parameter_str = ""
        # map method names to parameter names to include them in plot for comparison
        # some clustering methods can have multiple parameters
        map_method_parameter = {PeakAlignmentMethod.AFFINITY_PROPAGATION_CLUSTERING.name: "preference",
                                PeakAlignmentMethod.MEAN_SHIFT_CLUSTERING.name: "bandwidth",
                                PeakAlignmentMethod.K_MEANS_CLUSTERING.name: "n_clusters",
                                PeakAlignmentMethod.DB_SCAN_CLUSTERING.name: "eps",
                                PeakAlignmentMethod.WINDOW_FRAME.name: "distance_threshold",
                                PeakAlignmentMethod.PROBE_CLUSTERING.name: "threshold_inverse_reduced_mobility",
                                # threshold_inverse_reduced_mobility = threshold_inverse_reduced_mobility,
                                # threshold_scaling_retention_time = threshold_scaling_retention_time
                                }
        method_parameter = param_dict.get(map_method_parameter[method_name], '')
        method_parameter_name = map_method_parameter[method_name]

        if param_dict:
            if isinstance(method_parameter, float):
                method_parameter_str = "{:.4f}".format(method_parameter)
            elif isinstance(method_parameter, tuple):
                method_parameter_str = " ".join("{:.4f}").format(*method_parameter)
            else:
                method_parameter_str = str(method_parameter)
            return "{} = {}".format(method_parameter_name, method_parameter_str)
        return method_parameter_str

    @staticmethod
    def _plot_clustering(intermediate_analysis_result, plot_parameters):
        """
        Plot Peak alignment result 1 png per peak_detection method, using scatterplot
        :param intermediate_analysis_result: MccImsAnalysis object
        :param plot_parameters:
        :return:
        """
        figure_dir = f"{plot_parameters['plot_dir']}clustering"
        # first check whether directory exists for figure output - or whether we want to use a buffer
        if not plot_parameters.get('use_buffer', False):
            Path(figure_dir).mkdir(parents=True, exist_ok=True)

        plots_to_return = []
        peak_alignment_result = intermediate_analysis_result.peak_alignment_result
        peak_detection_steps = intermediate_analysis_result.peak_detection_combined
        clustering_parameters = peak_alignment_result.params_used_for_clustering
        n_clusters_dict = peak_alignment_result.number_of_clusters

        clustering_method_name = peak_alignment_result.params_used_for_clustering['peak_alignment_step'].name

        for peak_detection in peak_detection_steps:
            # map method names to parameter names to include them in plot for comparison
            method_parameter_str = ClusterPlot._map_clustering_parameters_to_string(param_dict=clustering_parameters,
                                                                                    method_name=clustering_method_name)
            title_string = "Method = {}, {} clusters, {}".format(
                clustering_method_name, n_clusters_dict[peak_detection.name].shape[0], method_parameter_str)

            fig_name = "{}clustering/{}_{}_{}_{}.png".format(
                plot_parameters['plot_dir'], plot_parameters['plot_prefix'], peak_detection.name, clustering_method_name,
                method_parameter_str)
            # x = peak_alignment_result.peak_coordinates[peak_detection.name]['inverse_reduced_mobility']
            # y = peak_alignment_result.peak_coordinates[peak_detection.name]['retention_time']
            # plt.scatter(x, y)
            coordinate_df = peak_alignment_result.peak_coordinates[peak_detection.name]
            # Do we need to use a consistent axis range? - not really
            sns.set_style('whitegrid')
            g = sns.lmplot(data=coordinate_df, x='inverse_reduced_mobility', y='retention_time', fit_reg=False)
            ax = g.axes[0, 0]
            ax.margins(0.05)

            # Compose multiple plots to one pdf file
            # https: // matplotlib.org / faq / howto_faq.html  # save-multiple-plots-to-one-pdf-file
            xlabel = "Inverse Reduced Mobility (1/k0) [Vs/cm^2]"
            ylabel = "Retention Time [s]"

            plt.yticks(rotation=0)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            ax.set_title(title_string)

            fig = plt.gcf()
            # distinguish different used methods by using tuples / a hashmap?
            plots_to_return.append((clustering_method_name, peak_detection.name, save_plot_to_buffer(plot_parameters, fig), fig_name))
            if not plot_parameters.get('use_buffer', False):
                fig.savefig(fig_name, dpi=300, bbox_inches='tight', format="png")
            plt.close()

        return plots_to_return

    @staticmethod
    def _plot_multiple_peak_detection_methods_clustering(mcc_ims_analysis, plot_parameters):
        """
        For each Peak alignment results plot 1 png
        :param mcc_ims_analysis: MccImsAnalysis object
        :param plot_parameters:
        :return:
        """
        figure_dir = f"{plot_parameters['plot_dir']}clustering"
        # first check whether directory exists for figure output - or whether we want to use a buffer
        if not plot_parameters.get('use_buffer', False):
            Path(figure_dir).mkdir(parents=True, exist_ok=True)

        peak_alignment_result = mcc_ims_analysis.peak_alignment_result
        peak_detection_steps = mcc_ims_analysis.peak_detection_combined
        clustering_parameters = peak_alignment_result.params_used_for_clustering
        # n_clusters_dict = peak_alignment_result.number_of_clusters

        method = peak_alignment_result.params_used_for_clustering['peak_alignment_step'].name

        method_parameter_str = ClusterPlot._map_clustering_parameters_to_string(param_dict=clustering_parameters,
                                                                                method_name=method)
        title_string = "Method = {}, {}".format(
            method, method_parameter_str)

        fig_name = "{}clustering/{}_{}_{}_{}.png".format(
            plot_parameters['plot_dir'], plot_parameters['plot_prefix'], "multi_peak_detections", method,
            method_parameter_str)

        # assign pdm column for hue
        # join them together in dataframe
        # use lmplot and hue=pdm
        df = pd.concat((peak_coords.assign(peak_detection_method=pdm) for pdm, peak_coords in peak_alignment_result.peak_coordinates.items()))
        sns.set_style('whitegrid')
        g = sns.lmplot(data=df, x='inverse_reduced_mobility', y='retention_time', fit_reg=False,
                       hue='peak_detection_method')
        ax = g.axes[0, 0]
        ax.margins(0.05)

        # Compose multiple plots to one pdf file
        # https: // matplotlib.org / faq / howto_faq.html  # save-multiple-plots-to-one-pdf-file
        xlabel = "Inverse Reduced Mobility (1/k0) [Vs/cm^2]"
        ylabel = "Retention Time [s]"

        plt.yticks(rotation=0)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(title_string)

        fig = plt.gcf()
        # distinguish different used methods by using tuples / a hashmap?
        if not plot_parameters.get('use_buffer', False):
            fig.savefig(fig_name, dpi=300, bbox_inches='tight', format="png")
        plot_to_return = save_plot_to_buffer(plot_parameters, fig)
        plt.close()
        return plot_to_return, fig_name

    @staticmethod
    def _compute_stepwidth_reflecting_multiple(vals, multiple):
        stepwidth = np.mean([vals[i+1] - vals[i] for i in range(len(vals) - 1)])
        return (multiple / float(stepwidth))


    @staticmethod
    def compute_stepwidth_fast(vals, target_size, num_points_axis, start_at=0.0):
        if start_at - 0.001 < 0.0:
            return target_size / (vals[-1]/num_points_axis)
        else:
            # print(f"Using cropped adjustment with {start_at}")
            # need to adjust stepwidth - as we potentially scaled a cropped image
            return target_size / ((vals[-1]-start_at) / num_points_axis)

    #
    # @staticmethod
    # def _translate_values_to_indices(index_values, to_scale, use_crop_correction=False):
    #     ids = np.asarray(index_values, dtype=float)
    #     ids_stepwidth = np.mean([ids[i+1] - ids[i] for i in range(len(ids) - 1)])
    #
    #     # get lowest val and scale by stepwidth
    #     lowest_val = np.min(ids)
    #     if use_crop_correction:
    #         scaled_lowest_val = lowest_val // ids_stepwidth + (lowest_val % ids_stepwidth) / ids_stepwidth
    #     else:
    #         scaled_lowest_val = 0.0
    #
    #     # need to assign index - get approx index and use float as relative index to scale
    #     # remove auto cropped out index (by heatmapplot) from calculated index
    #     def scale_to_index(val):
    #         basic_index = val // ids_stepwidth
    #         rest_index = (val % ids_stepwidth) / ids_stepwidth
    #         return (basic_index + rest_index) - scaled_lowest_val
    #
    #     return np.vectorize(scale_to_index)(to_scale)


    @staticmethod
    def _add_scaled_coords_fast(irm_index_values, rt_index_values, coords):
        """
        Scale coords using index values - and add with "scaled_" prefix to coords DataFrame
        :param irm_index_values:
        :param rt_index_values:
        :param coords:
        :return:
        """
        # for irm we need to crop correct -> need to get lowest value from coord axis and subtract the scaled value from
        # all our values, otherwise the coords will be right shifted
        irm_ids = np.asarray(irm_index_values, dtype=float)

        # compute width in pixels that corresponds to a change in 1.0 on the axis


        # irm_to_1_factor = 1/(irm_ids[-1]/2500.0)
        # irm_range = irm_ids[-1] - irm_ids[0]
        min_irm_val = max(0.0, np.min(irm_ids))
        irm_to_1_factor = 1/((irm_ids[-1] - min_irm_val) / 2500.0)

        pixels_to_shift_left = min_irm_val * irm_to_1_factor
        # print(f"Shifting rects by {pixels_to_shift_left} to the left")

        rt_ids = np.asarray(rt_index_values, dtype=float)
        rt_to_1_factor = 1 / (rt_ids[-1] / 2500)
        # rt_to_1_factor pixels is a change in one second retention time

        new_coords = coords.assign(
            scaled_irm=(coords['inverse_reduced_mobility']*irm_to_1_factor) - pixels_to_shift_left,
            scaled_rt=coords['retention_time']*rt_to_1_factor,
            scaled_irm_radius=coords['radius_inverse_reduced_mobility']*irm_to_1_factor,
            scaled_rt_radius=coords['radius_retention_time']*rt_to_1_factor)
        return new_coords



    @staticmethod
    def setup_axes_for_fast_intensity_matrix(ax, irm_vals, rt_vals, y_steps_width=25.0, x_steps_width=0.1,
                                             xlabel="Inverse Reduced Mobility (1/k0) [Vs/cm^2]",
                                             ylabel="Retention Time [s]", x_size=2500, y_size=2500):
        # need to adjust tick positions to multiples of 25 and 0.1 otherwise - its going to use the index of the df
        ax.yaxis.set_major_locator(ticker.MultipleLocator(
            ClusterPlot.compute_stepwidth_fast(rt_vals, y_steps_width, y_size)))

        ax.xaxis.set_major_locator(ticker.MultipleLocator(
            ClusterPlot.compute_stepwidth_fast(irm_vals, x_steps_width, x_size, start_at=irm_vals[0])))

        rt_top_tick = ((np.max(rt_vals) // y_steps_width) + 2.0) * y_steps_width

        irm_top_tick = ((np.max(irm_vals) // x_steps_width) + 2.0) * x_steps_width
        # get one ticklable bellow the actual one, as the first one gets swallowed
        irm_bottom_tick = max(0.0, ((np.min(irm_vals) // x_steps_width) - 1.0) * x_steps_width)

        # need to add 0.0 twice to y-axis as it "swallows" the first 0.0
        y_tick_labels = [0.0]
        for ele in np.arange(0, rt_top_tick, y_steps_width):
            y_tick_labels.append(ele)

        # for x ticks we have another problem, as we crop a significant part of the axis,
        #   our limit is not 0.0, but usually 0.4
        # need to use continous array and add the minimum value
        if irm_bottom_tick == 0.0:
            x_tick_labels = [0.0]
            x_tick_labels.extend(np.arange(0.0, irm_top_tick, x_steps_width))
        else:
            x_tick_labels = np.ascontiguousarray(np.arange(0, irm_top_tick, x_steps_width)) + irm_bottom_tick

        # need to convert to rounded representation
        x_tick_labels = np.round(x_tick_labels, decimals=2)
        # now we assign the "multiple of" labels to the axis - Fixed
        ax.yaxis.set_major_formatter(ticker.FixedFormatter(y_tick_labels))
        ax.xaxis.set_major_formatter(ticker.FixedFormatter(x_tick_labels))
        # reduce font size for labels
        ax.xaxis.set_tick_params(labelsize=7)
        ax.yaxis.set_tick_params(labelsize=7)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        #  remove black box surrounding the canvas at top and right side
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)


    @staticmethod
    def _setup_overlay_fast(intensity_matrix, plot_parameters, figure_dir):
        # filename = mcc_ims_measurement.filename
        fig, ax = HeatmapPlot.prepare_fast_heatmap_plot(intensity_matrix, {"cmap":plot_parameters['colormap'], "vmin":0, "vmax":1})

        # to plot we need to use the indices of the intensity matrix - not the actual values
        irm_vals = np.asarray(intensity_matrix.index.values, dtype=float)
        rt_vals = np.asarray(intensity_matrix.T.index.values, dtype=float)

        if not plot_parameters.get('use_buffer', False):
            Path(figure_dir).mkdir(parents=True, exist_ok=True)

        return ax, fig, irm_vals, rt_vals

    @staticmethod
    def _construct_coord_rects(scaled_coords, color):
        # lower left edge is at x,y - so we need to remove radius to match it
        # (x,y), w,h
        rect_coord_tuples = [((row['scaled_irm'] - row['scaled_irm_radius'],
           row['scaled_rt'] - row['scaled_rt_radius']),
          2. * row['scaled_irm_radius'], 2. * row['scaled_rt_radius'],
          )
                for _, row in scaled_coords.iterrows()]
        # print(f"Passed color = {color}")
        return [patches.Rectangle(*coord, fill=False, edgecolor=color, linestyle='dashed', )
                for coord in rect_coord_tuples]


    @staticmethod
    def _plot_overlay_clustering_helper_fast(intensity_matrix, pdm_coord_tuples, plot_parameters, title_dict, best_features_df=None):
        """
        Plot intensity_matrix and detected peaks on top with deviation indicated with rects
        iterating over multiple peak_detection results
        :param intensity_matrix:
        :param pdm_coord_tuples:
        :param plot_parameters:
        :param title_dict:
        :param best_features_df:
        :return:
        """
        return_figures = []
        figure_dir = f"{plot_parameters['plot_dir']}overlay/"
        # first check whether directory exists for figure output - or whether we want to use a buffer
        if not plot_parameters.get('use_buffer', False):
            Path(figure_dir).mkdir(parents=True, exist_ok=True)

        # plot intensity matrix and setup axis
        ax, fig, irm_vals, rt_vals = ClusterPlot._setup_overlay_fast(intensity_matrix, plot_parameters, figure_dir)

        if best_features_df is not None:
            # CASE1 - Best Overlay Plot
            for pdmn, peak_coords_original in pdm_coord_tuples:
                title_dict['peak_detection_method_name'] = pdmn
                best_peaks_by_pdm = best_features_df[best_features_df['peak_detection_method_name'] == pdmn]

                # scaled_coords_original = ClusterPlot._add_scaled_coords(irm_vals, rt_vals, peak_coords_original)

                # now iterate over different Feature reduction methods
                for performance_measure_name, best_peaks_by_pdm_pm in best_peaks_by_pdm.groupby("performance_measure_name"):
                    title_dict['performance_measure'] = performance_measure_name
                    # TODO - also group reduced features by class comparison? - so far only binary support
                    best_peak_coords = best_peaks_by_pdm_pm[peak_coords_original.columns.values]
                    scaled_best_coords = ClusterPlot._add_scaled_coords_fast(irm_vals, rt_vals, best_peak_coords)
                    # print(scaled_coords_original)
                    # print(scaled_best_coords)
                    # TO check if we have identical peak_coords, call this:
                    # print(scaled_coords_original.loc[scaled_coords_original['peak_id'].isin(scaled_best_coords.peak_id.values)].reset_index(inplace=True) == scaled_best_coords.reset_index(inplace=True))
                    # rects_orig = ClusterPlot._construct_coord_rects(scaled_coords_original, color="red")
                    rects_best = ClusterPlot._construct_coord_rects(scaled_best_coords, color="blue")
                    added_rects = [ax.add_patch(rect) for rect in rects_best]

                    title_string = f"Best Peaks in {title_dict['measurement_name']},\n {title_dict['peak_detection_method_name']}, {title_dict['performance_measure']}"

                    ax.set_title(title_string)
                    prefix = ""
                    if title_dict['dataset_name']:
                        prefix = f"{title_dict['dataset_name']}_"
                    figure_name = f"{prefix}best_overlay_{title_dict['peak_detection_method_name']}_{title_dict['measurement_name']}_{title_dict['performance_measure']}.png"

                    if not plot_parameters.get('use_buffer', False):
                        fig.savefig(f"{figure_dir}{figure_name}", dpi=300, bbox_inches='tight', format="png")
                    return_figures.append(((pdmn, title_dict['measurement_name'], performance_measure_name), save_plot_to_buffer(plot_parameters, fig), figure_name))
                    # remove rects for next iteration
                    for rect in added_rects: rect.remove()

        else:
            # CASE2 - Overlay Plot by PeakDetection
            # dont plot best features - only peak detection
            for pdmn, peak_coords in pdm_coord_tuples:
                title_dict['peak_detection_method_name'] = pdmn
                # are the coords correct?
                scaled_coords = ClusterPlot._add_scaled_coords_fast(irm_vals, rt_vals, peak_coords)

                rects = ClusterPlot._construct_coord_rects(scaled_coords, color="black")
                # cluster_centers = [(row['inverse_reduced_mobility'], row['retention_time']) for _, row in new_scaled_coords.iterrows()]
                added_rects = [ax.add_patch(rect) for rect in rects]

                # Scatterplot to mark peak positions
                # ax.scatter(data=scaled_coords, x="scaled_irm", y="scaled_rt", marker='*')
                title_string = f"Peaks in {title_dict['measurement_name']}, {title_dict['peak_detection_method_name']}"

                ax.set_title(title_string)
                prefix = ""
                if title_dict['dataset_name']:
                    prefix = f"{title_dict['dataset_name']}_"
                figure_name = f"{prefix}overlay_{title_dict['peak_detection_method_name']}_{title_dict['measurement_name']}.png"

                if not plot_parameters.get('use_buffer', False):
                    fig.savefig(f"{figure_dir}{figure_name}", dpi=300, bbox_inches='tight', format="png" )
                return_figures.append((pdmn, title_dict['measurement_name'], save_plot_to_buffer(plot_parameters, fig), figure_name))
                # remove rects for next iteration
                for rect in added_rects: rect.remove()
        plt.close()

        return return_figures

    @staticmethod
    def _plot_overlay_best_features_aligned_classwise(mcc_ims_analysis, plot_parameters):
        """
        Plot classwise intensity_matrix and detected peaks on top with deviation indicated with seethrough rects - best features in blue
        :param mcc_ims_analysis:
        :param plot_parameters:
        :return:
        """
        print("Plotting Overlay Best Features Plots")
        collected_figure_tuples = []
        peak_alignment_result = mcc_ims_analysis.peak_alignment_result

        title_dict = {'dataset_name': mcc_ims_analysis.dataset_name}

        class_labels = np.unique([m.class_label for m in mcc_ims_analysis.measurements])
        class_label_dict = {cl: [] for cl in class_labels}
        for m in mcc_ims_analysis.measurements:
            class_label_dict[m.class_label].append(m)

        # create consensus df for each class
        consensus_dfs = {}
        for class_label, measurements in class_label_dict.items():
            consensus_dfs[class_label] = HeatmapPlot.compute_averaged_df([m.df for m in measurements])

        peak_alignment_result = mcc_ims_analysis.peak_alignment_result

        for class_label, consensus_df in consensus_dfs.items():
            # title_dict['class_label'] = class_label
            class_str = "best_by_class-{}".format(class_label.replace(" ", ""))
            print("class_str={}".format(class_str))
            print("class_label={}".format(class_label))
            title_dict['measurement_name'] = class_str
            # speed up by iterating over measurement, updating rects and keeping heatmap
            collected_figure_tuples.extend(
                (ClusterPlot._plot_overlay_clustering_helper_fast(
                    intensity_matrix=consensus_df, pdm_coord_tuples=peak_alignment_result.peak_coordinates.items(),
                    plot_parameters=plot_parameters, title_dict=title_dict, best_features_df=mcc_ims_analysis.analysis_result.best_features_df))
            )
        return collected_figure_tuples

    @staticmethod
    def _plot_overlay_best_features(mcc_ims_analysis, plot_parameters):
        """
        Plot intensity_matrix and detected peaks on top with deviation indicated with seethrough rects - best features in blue
        :param mcc_ims_analysis:
        :param plot_parameters:
        :return:
        """
        print("Plotting Overlay Best Features Plots")
        collected_figure_tuples = []
        # could compute intersection of all and best features
        # using abstract function for adding peak coords
        peak_alignment_result = mcc_ims_analysis.peak_alignment_result
        title_dict = {'dataset_name': mcc_ims_analysis.dataset_name}
        for m in mcc_ims_analysis.measurements:
            parsed_filename = "{}_ims.csv".format(m.filename.split("_ims")[0])
            title_dict['measurement_name'] = parsed_filename
            # speed up by iterating over measurement, updating rects and keeping heatmap
            collected_figure_tuples.extend(
                (ClusterPlot._plot_overlay_clustering_helper_fast(
                    intensity_matrix=m.df, pdm_coord_tuples=peak_alignment_result.peak_coordinates.items(),
                    plot_parameters=plot_parameters, title_dict=title_dict, best_features_df=mcc_ims_analysis.analysis_result.best_features_df))
            )
        return collected_figure_tuples


    @staticmethod
    def _plot_overlay_aligned_clustering(mcc_ims_analysis, plot_parameters):
        """
        Plot intensity_matrix and detected peaks on top (red rects)
        :param mcc_ims_analysis:
        :param plot_parameters:
        :return:
        """
        print("Plotting Overlay Plots")
        collected_figure_tuples = []
        peak_alignment_result = mcc_ims_analysis.peak_alignment_result
        title_dict = {'dataset_name': mcc_ims_analysis.dataset_name}
        for m in mcc_ims_analysis.measurements:

            # remove double .csv and _preprocessed in name

            preprocessed_filename = m.filename.split(".csv.csv")[0]
            preprocessed_filename = preprocessed_filename.split("_preprocessed")[0]

            title_dict['measurement_name'] = preprocessed_filename
            # speed up by iterating over measurement, updating rects and keeping heatmap
            collected_figure_tuples.extend(
                    (ClusterPlot._plot_overlay_clustering_helper_fast(
                    intensity_matrix=m.df, pdm_coord_tuples=peak_alignment_result.peak_coordinates.items(),
                    plot_parameters=plot_parameters, title_dict=title_dict))
                )
        return collected_figure_tuples


    @staticmethod
    def _plot_overlay_aligned_clustering_classwise(mcc_ims_analysis, plot_parameters):
        """
        Plot intensity_matrix and detected peaks on top of averaged measurement by class_label (red rects)
        :param mcc_ims_analysis:
        :param plot_parameters:
        :return:
        """
        print("Plotting Classwise Overlay Plots")
        collected_figure_tuples = []
        peak_alignment_result = mcc_ims_analysis.peak_alignment_result

        title_dict = {'dataset_name': mcc_ims_analysis.dataset_name}

        class_labels = np.unique([m.class_label for m in mcc_ims_analysis.measurements])
        class_label_dict = {cl: [] for cl in class_labels}
        for m in mcc_ims_analysis.measurements:
            class_label_dict[m.class_label].append(m)

        # create consensus df for each class
        consensus_dfs = {}
        for class_label, measurements in class_label_dict.items():
            consensus_dfs[class_label] = HeatmapPlot.compute_averaged_df([m.df for m in measurements])

        for class_label, consensus_df in consensus_dfs.items():
            # title_dict['class_label'] = class_label
            class_str = "by_class-{}".format(class_label.replace(" ",""))
            print("class_str={}".format(class_str))
            print("class_label={}".format(class_label))
            title_dict['measurement_name'] = class_str
            # speed up by iterating over measurement, updating rects and keeping heatmap
            collected_figure_tuples.extend(
                    (ClusterPlot._plot_overlay_clustering_helper_fast(
                    intensity_matrix=consensus_df, pdm_coord_tuples=peak_alignment_result.peak_coordinates.items(),
                    plot_parameters=plot_parameters, title_dict=title_dict))
                )
        return collected_figure_tuples

    ClusterBasic = _plot_clustering
    ClusterMultiple = _plot_multiple_peak_detection_methods_clustering
    OverlayClasswiseAlignment = _plot_overlay_aligned_clustering_classwise
    OverlayAlignment = _plot_overlay_aligned_clustering
    OverlayBestFeaturesAlignment = _plot_overlay_best_features
    OverlayBestFeaturesClasswiseAlignment = _plot_overlay_best_features_aligned_classwise


class FeatureMatrixPlot(object):
    """
    Collection of Feature matrix visualizations
    """

    @staticmethod
    def _plot_patients_vs_features(df, index_label_mapping, measurement_mapping, plot_parameters):
        """
        Plot heatmap of patient csv_names vs peaks,
        :param df: a matrix with encoded features
        :type df: pd.Dataframe
        :param index_label_mapping:
        :return:
        """
        plot_prefix = ''

        if 'plot_prefix' in plot_parameters:
            plot_prefix = plot_parameters['plot_prefix']

        # print(m_train.head())
        # TODO make plot generic for multi label
        # save class labels independent from names of classes
        class_labels = np.unique([a for a in index_label_mapping.values()])
        number_of_class_labels = len(class_labels)

        if number_of_class_labels > 2:
            print("Plot only available for two class labels, will use first two.")
        print(class_labels)
        print(number_of_class_labels)
        map_labels = {l: [] for l in class_labels}
        for k, v in index_label_mapping.items(): map_labels[v].extend([k]);

        # plot a heatmap
        plt.close('all')
        fig, axn = plt.subplots(2, 1, sharex=True, figsize=(9, 4))
        bw_colormap = ListedColormap(['black', 'white'])
        for i, ax in enumerate(axn.flat):
            if i == 0:
                # sns.heatmap(df.loc[citrus_labels], yticklabels=citrus_yticks, xticklabels=15, cbar=False, ax=ax)
                # sns.heatmap(df.loc[map_labels[class_labels[0]]], yticklabels=[measurement_mapping[index_number] for index_number in map_labels[class_labels[0]]], xticklabels=15, cbar=False, ax=ax)
                sns.heatmap(df.loc[map_labels[class_labels[0]]],
                            yticklabels=[measurement_mapping[index_number] for index_number in map_labels[class_labels[0]]],
                            xticklabels=15, cmap=bw_colormap, ax=ax, cbar=False)
                ax.set_ylabel("citrus")
                ax.set_title('Patients vs Features (Peaks)')
            else:
                # sns.heatmap(df.loc[original_labels], yticklabels=original_yticks, xticklabels=15, cbar=False, ax=ax)
                sns.heatmap(df.loc[map_labels[class_labels[1]]],
                            yticklabels=[measurement_mapping[index_number] for index_number in map_labels[class_labels[1]]],
                            xticklabels=15, cmap=bw_colormap, cbar=False, ax=ax)
                ax.set_ylabel("original")
            plt.sca(ax)
            plt.yticks(rotation=0)

        # set title
        # fig.subtitle('Features (Peaks) vs Patients')
        # configure legend
        black_patch = matplotlib.patches.Patch(color='black', label='0')
        white_patch = matplotlib.patches.Patch(color='white', label='1')
        plt.legend(title='Peak present', handles=[black_patch, white_patch], bbox_to_anchor=(1.2, 1.3))
        # fig.tight_layout()
        plt.savefig("%sheatmaps/%s_patients_vs_peaks.png" % (plot_parameters['plot_dir'], plot_prefix), bbox_inches='tight')
        # plt.close('all')
        # # sns.heatmap(m_train[original_labels], ax=ax)
        #
        # # print(m_train)
        # # print(feature_index)
        # # print(feature_index)
        plt.close('all')


    BitMatrix = _plot_patients_vs_features


class MaskPlot(object):
    """
    Collection of mask visualizations after peak detection
    """

    @staticmethod
    def _plot_mask_from_peak_detection_result(peak_detection_result, plot_parameters, ):
        """
        Visualize peak detection result of PeakDetectionResult
        :type peak_detection_result: model.BreathCore.PeakDetectioResut
        :param peak_detection_result:
        :param plot_parameters:
        :return:
        """
        measurement_name = peak_detection_result.outname
        MaskPlot._plot_mask(peak_detection_result.peak_detection_mask, measurement_name, plot_parameters)

    @staticmethod
    def _plot_mask(intensity_matrix, measurement_name, plot_parameters):
        # TODO mention the peak detection method in a generic way - right now only available from TOPHAT
        print("Plotting mask from {} created with {}".format(measurement_name, "TopHat"))
        plot_prefix = ''
        if 'plot_prefix' in plot_parameters:
            plot_prefix = plot_parameters['plot_prefix']

        """Set the x- and y-axis labels"""
        # xmax = intensity_matrix['inverse_reduced_mobility'].max()
        # pdb.set_trace()
        inverse_reduced_mobility_vals = intensity_matrix.index.values
        retention_time_vals = intensity_matrix.columns.values[:-1]
        xstep = [str(n) for n in np.arange(np.around(inverse_reduced_mobility_vals.min(), decimals=1),
                                           inverse_reduced_mobility_vals.max(), 0.1)]
        # ymax = intensity_matrix['retention_time'].max()

        ystep = [str(ys) for ys in np.arange(0.0, retention_time_vals.max(), 25)]
        majorsx = xstep
        majorsy = ystep
        """Create the heatmap"""
        ax = sns.heatmap(intensity_matrix.T, cmap=plot_parameters['colormap'],
                         xticklabels=len(inverse_reduced_mobility_vals) // len(majorsx),
                         yticklabels=len(retention_time_vals) // len(majorsy))
        ax.set_xlabel('1/K0')
        ax.xaxis.set_major_formatter(ticker.FixedFormatter(majorsx))
        ax.yaxis.set_major_formatter(ticker.FixedFormatter(majorsy))
        plt.yticks(rotation=0)
        sns.set_style('ticks')
        # ax.set(xlabel="Retention Time [s]", ylabel="Inverse Reduced Mobility (1/R0) [Vs/cm^2]")
        ax.set_xlabel("Inverse Reduced Mobility (1/k0) [Vs/cm^2]")
        ax.set_ylabel("Retention Time [s]")
        ax.set_title('Measurement {}'.format(measurement_name))

        fig = ax.get_figure()
        fig.set_figsize = (9, 4)
        fig.savefig('{}masks/{}_mask_plot_{}.png'.format(plot_parameters['plot_dir'],
                                                     plot_prefix, measurement_name[:-4]),
                    bbox_inches='tight')
        fig.clf()
        plt.close('all')

    MaskHeatmap = _plot_mask_from_peak_detection_result


class RocCurvePlot(object):
    """
    Plot the ROC curve of the AnalysisResult
    """

    @staticmethod
    def _plot_two_class_roc_curve(analysis_result, plot_parameters, limit_to_peak_detection_method_name=False):
        from ..model.BreathCore import PerformanceMeasure
        return_figures = []
        print("Plotting binary ROC curve")
        # assert isinstance(analysis_result, AnalysisResult)
        plot_prefix = ''
        if 'plot_prefix' in plot_parameters:
            plot_prefix = plot_parameters['plot_prefix']

        use_buffer = plot_parameters.get('use_buffer', False)

        figure_dir = f"{plot_parameters['plot_dir']}roc_curve"
        # first check whether directory exists for figure output - or whether we want to use a buffer
        if not use_buffer:
            Path(figure_dir).mkdir(parents=True, exist_ok=True)

        # plt.plot(analysis_result.statistics["false_positive_rates_of_splits"],
        #          analysis_result.statistics["true_positive_rates_of_splits"], lw=1, alpha=0.3, )
        # based on https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc_crossval.html
        # and https://stackoverflow.com/questions/42901264/identical-auc-for-each-fold-in-cross-validation-roc-curves-python
        alignment_name = analysis_result.peak_alignment_step if analysis_result.peak_alignment_step is not None else ""
        for evaluation_method_name, statistics_by_pdm_name in analysis_result.analysis_statistics_per_evaluation_method_and_peak_detection.items():
            # need to check first whether we computed a roc measure - only the case for RF eval
            if evaluation_method_name in [PerformanceMeasure.RANDOM_FOREST_CLASSIFICATION.name]:

                for peak_detection_method_name, statistics in statistics_by_pdm_name.items():
                    if limit_to_peak_detection_method_name and limit_to_peak_detection_method_name !=peak_detection_method_name:
                        continue
                    else:
                        # clear current figure
                        fig = plt.gcf()
                        auc_measures = statistics['auc_measures']
                        # print(auc_measures.keys())

                        # need to save tprs in another way, as the splits might not be of equal sizes
                        mean_tpr = 0.0
                        mean_fpr = np.linspace(0, 1, 100)
                        # print("no of splits", len(auc_measures['tpr_of_splits_by_class'][0]))


                        # go over splits and draw fpr/tpr
                        for split_no, (tpr_per_split, fpr_per_split, auc_per_split) in enumerate(zip(
                                    auc_measures['tpr_of_splits'],
                                    auc_measures['fpr_of_splits'],
                                    auc_measures['auc_of_splits'])):

                            plt.plot(fpr_per_split, tpr_per_split, lw=1, alpha=0.3, label='ROC fold %d (AUC = %0.2f)' % (split_no, auc_per_split))
                            # tprs.append(interp(mean_fpr, fpr_per_split, tpr_per_split))
                            mean_tpr += interp(mean_fpr, fpr_per_split, tpr_per_split)
                            mean_tpr[0] = 0.0

                        # add random classification performance line
                        plt.plot([0, 1], [0, 1], linestyle='--', lw=2, color='r',
                                 label='Random', alpha=.8)

                        # splits are not always of equal size
                        mean_tpr /= len(auc_measures['tpr_of_splits'])
                        mean_tpr[-1] = 1.0

                        mean_auc = auc(mean_fpr, mean_tpr)

                        # add the average classification line
                        plt.plot(mean_fpr, mean_tpr, color='green',
                                 label=r'Mean ROC (AUC = {:.2f} $\pm$ {:.2f})'.format(mean_auc, auc_measures['std_auc']),
                                 lw=2, alpha=.8)

                        std_tpr = np.std(mean_tpr, axis=0)
                        tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
                        tprs_lower = np.maximum(mean_tpr - std_tpr, 0)
                        plt.fill_between(mean_fpr, tprs_lower, tprs_upper, color='grey', alpha=.2,
                                         label=r'$\pm$ 1 std. dev.')

                        # fig.add_subplot(ax)
                        RocCurvePlot._label_roc_axes(peak_detection_method_name)
                        fig = plt.gcf()

                        fig.set_figsize = (9, 4)

                        if use_buffer:
                            fig_name = f"{plot_prefix}_roc_curve_plot.png"
                        else:
                            fig_name = f"{plot_parameters['plot_dir']}roc_curve/{plot_prefix}_roc_curve_plot_{peak_detection_method_name}_{alignment_name}.png"

                        if not plot_parameters.get('use_buffer', False):
                            fig.savefig(fig_name, dpi=300, bbox_inches='tight', )

                            # ((peak_detection, model_of_class, best_feature), save_plot_to_buffer(plot_parameters, fig), fig_name))
                        return_figures.append((evaluation_method_name, peak_detection_method_name, save_plot_to_buffer(plot_parameters, fig), fig_name))
                        plt.close()
                    # fig.clf()
        return return_figures

    @staticmethod
    def _plot_multiclass_class_roc_curve(analysis_result, plot_parameters, limit_to_peak_detection_method_name=False):
        from ..model.BreathCore import PerformanceMeasure
        return_figures = []
        print("Plotting multiclass ROC curve")
        plot_prefix = ''
        if 'plot_prefix' in plot_parameters:
            plot_prefix = plot_parameters['plot_prefix']

        use_buffer = plot_parameters.get('use_buffer', False)

        figure_dir = f"{plot_parameters['plot_dir']}roc_curve"
        # first check whether directory exists for figure output - or whether we want to use a buffer
        if not use_buffer:
            Path(figure_dir).mkdir(parents=True, exist_ok=True)


        unique_class_labels = np.sort(np.unique(analysis_result.class_labels))
        for evaluation_method_name, statistics_by_pdm_name in analysis_result.analysis_statistics_per_evaluation_method_and_peak_detection.items():
            # need to check first whether we computed a roc measure - only the case for RF eval
            if evaluation_method_name in [PerformanceMeasure.RANDOM_FOREST_CLASSIFICATION.name]:

                for peak_detection, statistics in statistics_by_pdm_name.items():
                    if limit_to_peak_detection_method_name and limit_to_peak_detection_method_name !=peak_detection:
                        continue
                    else:

                        # clear current figure
                        fig = plt.gcf()
                        auc_measures = statistics['auc_measures']
                        # dont count 'macro' key as class - ist just the average
                        number_of_classes = len(unique_class_labels)

                        # add random classification performance line
                        plt.plot([0, 1], [0, 1], linestyle='--', lw=2, color='r', label='Random', alpha=.8)

                        for class_label in range(number_of_classes):

                            # need to save tprs in another way, as the splits might not be of equal sizes
                            mean_tpr = 0.0
                            mean_fpr = np.linspace(0, 1, 100)

                            for i, (tpr_per_split, fpr_per_split, auc_per_split) in enumerate(
                                    zip(auc_measures['tpr_of_splits_by_class'][class_label],
                                        auc_measures['fpr_of_splits_by_class'][class_label],
                                        auc_measures['auc_of_splits_by_class'][class_label])):

                                # plt.plot(fpr_per_split, tpr_per_split, lw=1, alpha=0.3,
                                #          label='ROC fold %d (AUC = %0.2f)' % (i, auc_per_split))
                                # tprs.append(interp(mean_fpr, fpr_per_split, tpr_per_split))
                                mean_tpr += interp(mean_fpr, fpr_per_split, tpr_per_split)
                                mean_tpr[0] = 0.0
                                # print("split = {}".format(i))

                            # average over splits
                            mean_tpr /= len(auc_measures['tpr_of_splits_by_class'][0])
                            mean_tpr[-1] = 1.0

                            mean_auc = auc(mean_fpr, mean_tpr)

                            # add the average classification line
                            plt.plot(mean_fpr, mean_tpr,
                                     label=r'Class {} macro ROC (AUC = {:.2f} $\pm$ {:.2f})'.format(
                                         unique_class_labels[class_label], mean_auc, auc_measures['std_auc']),
                                         # unique_class_labels[class_label], auc_measures['mean_auc'], auc_measures['std_auc']),
                                     lw=2, alpha=.8)

                            # plt.fill_between(mean_fpr, tprs_lower, tprs_upper, color='grey', alpha=.2,
                            #                  label=r'$\pm$ 1 std. dev.')
                        RocCurvePlot._label_roc_axes(peak_detection)
                        fig = plt.gcf()

                        pas_n = ""
                        if analysis_result.peak_alignment_step is not None:
                            pas_n = analysis_result.peak_alignment_step.name

                        fig_name = f"{plot_parameters['plot_dir']}roc_curve/{plot_prefix}_multi_roc_curve_plot_{peak_detection}_{pas_n}.png"
                        # print(fig_name)
                        if not plot_parameters.get('use_buffer', False):
                            fig.savefig(fig_name, dpi=300, bbox_inches='tight')

                        return_figures.append((evaluation_method_name, peak_detection, save_plot_to_buffer(plot_parameters, fig), fig_name))
                            # ((peak_detection, model_of_class, best_feature), save_plot_to_buffer(plot_parameters, fig), fig_name))
                        plt.close()
        return return_figures


    @staticmethod
    def _label_roc_axes(peak_detection):
        plt.xlim([-0.05, 1.05])
        plt.ylim([-0.05, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('ROC for {}'.format(peak_detection))
        plt.legend(loc="lower right")

    ROCCurve = _plot_two_class_roc_curve
    MultiClassROCCurve = _plot_multiclass_class_roc_curve

class VennDiagram(object):
    """
    Plot Venn diagrams.

    The venn import allows to create venn diagrams with more than 3 sets
    """

    @staticmethod
    def _plot_venn_diagram_from_peak_detection_result(comparison_peak_detection_result, plot_parameters):
        """
        Visualize the overlap of different PeakDetectionMethods
        :param peak_detection_result:
        :param plot_parameters:
        :return:
        """
        VennDiagram._plot_venn2_diagram(comparison_peak_detection_result, plot_parameters)
        VennDiagram._plot_venn3_diagram(comparison_peak_detection_result, plot_parameters)
        VennDiagram._plot_venn4_diagram(comparison_peak_detection_result, plot_parameters)

    @staticmethod
    def _plot_venn2_diagram(comparison_peak_detection_result, plot_parameters):
        if comparison_peak_detection_result.shape[0] > 2:
            raise AttributeError("Venn2Diagram requires 2 sets. You gave {} sets."
                                 "Choose Venn3- or Venn4Diagram, respectively".format(
                comparison_peak_detection_result.shape[0]))
        plot_prefix = ''
        if 'plot_prefix' in plot_parameters:
            plot_prefix = plot_parameters['plot_prefix']
        peak_detections = sorted(comparison_peak_detection_result.index.values)
        individual_1 = comparison_peak_detection_result.loc[peak_detections[0]]['individual']
        individual_2 = comparison_peak_detection_result.loc[peak_detections[1]]['individual']
        intersection_12 = max([comparison_peak_detection_result.loc[peak_detections[0]][peak_detections[1]],
                               comparison_peak_detection_result.loc[peak_detections[1]][peak_detections[0]]])
        fig = plt.figure()
        v = venn2(subsets=(individual_1, individual_2, intersection_12), set_labels=peak_detections)
        c = venn2_circles(subsets=(individual_1, individual_2, intersection_12), linestyle='dashed', linewidth=1, color="grey")
        comp_str = f"venn2_{comparison_peak_detection_result.index.values[0]}_{comparison_peak_detection_result.index.values[1]}"
        fig.savefig(f"{plot_parameters['plot_dir']}venn_diagram/{plot_prefix}_{comp_str}.png", bbox_inches='tight')
        plt.close()

    @staticmethod
    def _plot_venn3_diagram(comparison_peak_detection_result, plot_parameters):

        if comparison_peak_detection_result.shape[0] < 3:
            raise AttributeError("Venn3Diagram requires 3 sets. You gave {} sets."
                                 "Choose Venn2- or Venn3Diagram, respectively".format(comparison_peak_detection_result.shape[0]))
        if comparison_peak_detection_result.shape[0] > 3:
            raise AttributeError("Venn3Diagram requires 3 sets. You gave {} sets."
                                 "Choose Venn2- or Venn4Diagram, respectively".format(comparison_peak_detection_result.shape[0]))
        plot_prefix = ''
        if 'plot_prefix' in plot_parameters:
            plot_prefix = plot_parameters['plot_prefix']
        peak_detections = sorted(comparison_peak_detection_result.index.values)
        individual_1 = comparison_peak_detection_result.loc[peak_detections[0]]['individual']
        individual_2 = comparison_peak_detection_result.loc[peak_detections[1]]['individual']
        individual_3 = comparison_peak_detection_result.loc[peak_detections[2]]['individual']
        intersection_12 = max([comparison_peak_detection_result.loc[peak_detections[0]][peak_detections[1]],
                               comparison_peak_detection_result.loc[peak_detections[1]][peak_detections[0]]])
        intersection_13 = max([comparison_peak_detection_result.loc[peak_detections[0]][peak_detections[2]],
                               comparison_peak_detection_result.loc[peak_detections[2]][peak_detections[0]]])
        intersection_23 = max([comparison_peak_detection_result.loc[peak_detections[1]][peak_detections[2]],
                               comparison_peak_detection_result.loc[peak_detections[2]][peak_detections[1]]])
        intersection_all = comparison_peak_detection_result.loc[:]['all'].max()
        fig = plt.figure()
        v = venn3(subsets=(individual_1, individual_2, intersection_12, individual_3, intersection_13, intersection_23,
                           intersection_all), set_labels=peak_detections)
        c = venn3_circles(subsets=(individual_1, individual_2, intersection_12, individual_3, intersection_13,
                                   intersection_23, intersection_all), linestyle='dashed', linewidth=1, color="grey")
        comp_str = f"venn3_{comparison_peak_detection_result.index.values[0]}_{comparison_peak_detection_result.index.values[1]}_{comparison_peak_detection_result.index.values[2]}"
        fig.savefig(f"{plot_parameters['plot_dir']}venn_diagram/{plot_prefix}_{comp_str}.png",bbox_inches='tight')
        plt.close()

    @staticmethod
    def _plot_venn4_diagram(comparison_peak_detection_result, plot_parameters):
        from breathpy.view import venn

        if comparison_peak_detection_result.shape[0] < 4:
            raise AttributeError("Venn4Diagram requires 4 sets. You gave {} sets."
                                 "Choose Venn2- or Venn3Diagram, respectively".format(comparison_peak_detection_result.shape[0]))
        if comparison_peak_detection_result.shape[0] > 4:
            raise AttributeError("Venn4Diagram requires 4 sets. You gave {} sets."
                                 "A VennDiagram with more than 4 sets is to complex. Choose another from of visualization".format(comparison_peak_detection_result.shape[0]))
        plot_prefix = ''
        if 'plot_prefix' in plot_parameters:
            plot_prefix = plot_parameters['plot_prefix']
        labels = dict()
        peak_detections = sorted(comparison_peak_detection_result.index.values)
        labels['1000'] = "{}".format(comparison_peak_detection_result.loc[peak_detections[0]]['individual'])
        labels['0100'] = "{}".format(comparison_peak_detection_result.loc[peak_detections[1]]['individual'])
        labels['0010'] = "{}".format(comparison_peak_detection_result.loc[peak_detections[2]]['individual'])
        labels['0001'] = "{}".format(comparison_peak_detection_result.loc[peak_detections[3]]['individual'])
        labels['1100'] = "{}".format(max([comparison_peak_detection_result.loc[peak_detections[0]][peak_detections[1]],
                                    comparison_peak_detection_result.loc[peak_detections[1]][peak_detections[0]]]))
        labels['1010'] = "{}".format(max([comparison_peak_detection_result.loc[peak_detections[0]][peak_detections[2]],
                                    comparison_peak_detection_result.loc[peak_detections[2]][peak_detections[0]]]))
        labels['1001'] = "{}".format(max([comparison_peak_detection_result.loc[peak_detections[0]][peak_detections[3]],
                                    comparison_peak_detection_result.loc[peak_detections[3]][peak_detections[0]]]))
        labels['0110'] = "{}".format(max([comparison_peak_detection_result.loc[peak_detections[1]][peak_detections[2]],
                                    comparison_peak_detection_result.loc[peak_detections[2]][peak_detections[1]]]))
        labels['0101'] = "{}".format(max([comparison_peak_detection_result.loc[peak_detections[1]][peak_detections[3]],
                                    comparison_peak_detection_result.loc[peak_detections[3]][peak_detections[1]]]))
        labels['0011'] = "{}".format(max([comparison_peak_detection_result.loc[peak_detections[2]][peak_detections[3]],
                                    comparison_peak_detection_result.loc[peak_detections[3]][peak_detections[2]]]))
        labels['1110'] = "{}".format(max([comparison_peak_detection_result.loc[peak_detections[0]]["{}_{}".format(peak_detections[1], peak_detections[2])],
                                    comparison_peak_detection_result.loc[peak_detections[1]]["{}_{}".format(peak_detections[0], peak_detections[2])],
                                    comparison_peak_detection_result.loc[peak_detections[2]]["{}_{}".format(peak_detections[0], peak_detections[1])]]))
        labels['1101'] = "{}".format(max([comparison_peak_detection_result.loc[peak_detections[0]]["{}_{}".format(peak_detections[1], peak_detections[3])],
                                    comparison_peak_detection_result.loc[peak_detections[1]]["{}_{}".format(peak_detections[0], peak_detections[3])],
                                    comparison_peak_detection_result.loc[peak_detections[3]]["{}_{}".format(peak_detections[0], peak_detections[1])]]))
        labels['1011'] = "{}".format(max([comparison_peak_detection_result.loc[peak_detections[0]]["{}_{}".format(peak_detections[2], peak_detections[3])],
                                    comparison_peak_detection_result.loc[peak_detections[2]]["{}_{}".format(peak_detections[0], peak_detections[3])],
                                    comparison_peak_detection_result.loc[peak_detections[3]]["{}_{}".format(peak_detections[0], peak_detections[2])]]))
        labels['0111'] = "{}".format(max([comparison_peak_detection_result.loc[peak_detections[1]]["{}_{}".format(peak_detections[2],peak_detections[3])],
                                    comparison_peak_detection_result.loc[peak_detections[2]]["{}_{}".format(peak_detections[1], peak_detections[3])],
                                    comparison_peak_detection_result.loc[peak_detections[3]]["{}_{}".format(peak_detections[1], peak_detections[2])]]))
        labels['1111'] = "{}".format(comparison_peak_detection_result.loc[:]['all'].max())
        fig, ax = venn.venn4(labels=labels, names=peak_detections)
        fig.savefig(f"{plot_parameters['plot_dir']}venn_diagram/{plot_prefix}_venn4.png", bbox_inches='tight')
        plt.close()

    Venn2Diagram = _plot_venn2_diagram
    Venn3Diagram = _plot_venn3_diagram
    Venn4Diagram = _plot_venn4_diagram


class BoxPlot(object):
    """
    Create Box plots
    """

    @staticmethod
    def convert_pval(n):
        return '%.1E' % Decimal(n)

    @staticmethod
    def _plot_boxplot_of_best_features_seaborn(analysis_result, plot_parameters, limit_to_peak_detection_method_name=False):
        from ..model.BreathCore import PerformanceMeasure
        from ..model.BreathCore import MccImsAnalysis
        import statannot
        sns.set_style("whitegrid")
        # pdb.set_trace()
        # types dont match when using in web, as we use different imports
        # assert isinstance(analysis_result, AnalysisResult)
        plots_to_return = []
        print("Plotting Boxplots of the best features SEABORN")
        figure_dir = f"{plot_parameters['plot_dir']}boxplots"
        # first check whether directory exists for figure output - or whether we want to use a buffer
        if not plot_parameters.get('use_buffer', False):
            Path(figure_dir).mkdir(parents=True, exist_ok=True)

        corrected_pvalues_df = analysis_result.get_pvalues_df(
            n_features=MccImsAnalysis.DEFAULT_PERFORMANCE_MEASURE_PARAMETERS[PerformanceMeasure.FDR_CORRECTED_P_VALUE][
                'n_of_features'])


        for peak_detection_name, peak_intensities_by_class_by_peak in analysis_result.peak_intensities_by_pdm_by_class_by_peak.items():
            if limit_to_peak_detection_method_name and peak_detection_name != peak_detection_name:
                plt.close()
                continue
            else:
                corrected_pvals_by_pdmn = corrected_pvalues_df[
                        corrected_pvalues_df['peak_detection_method_name'] == peak_detection_name]
                #
                # classes = peak_intensities_by_class_by_peak.keys()
                #
                # # classes have the same peak_ids
                # peak_ids = sorted(peak_intensities_by_class_by_peak[classes[0]].keys())
                best_feature_set = set(np.unique(corrected_pvals_by_pdmn['peak_id']))
                records = []
                for model_of_class, intensities_by_pid in peak_intensities_by_class_by_peak.items():
                    for peak_id, intensities in intensities_by_pid.items():

                        # only create record if peak_id in best_features - otherwise we get too many plots and it takes forever
                        if peak_id in best_feature_set:
                            record_dict = {}
                            record_dict['class_str'] = model_of_class
                            record_dict['peak_id'] = peak_id
                            record_dict['intensities'] = intensities
                            records.append(record_dict)
                df = pd.DataFrame.from_records(records)

                # classes = np.unique(df['class_str'].values)
                # peak_ids = np.unique(df['peak_id'].values)

                for peak_id, rows in df.groupby('peak_id'):
                    # pdb.set_trace()
                    list_of_intensities_of_classes = [*rows['intensities'].values]
                    class_strs = [*rows['class_str'].values]

                    #  make seaborn boxplot here
                    fig = plt.figure()
                    ax = fig.add_subplot(111)
                    # need to unpack intensities for seaborn
                    unpacked_df_lis = []
                    for class_str, class_rows in rows.groupby('class_str'):
                        data = {"Intensity": class_rows.intensities.values[0],
                                "class": [class_str]*len(class_rows.intensities.values[0])}
                        unpacked_df_lis.append(pd.DataFrame(data))
                    unpacked_df = pd.concat(unpacked_df_lis)

                    # ax = sns.boxplot(data=unpacked_df, x='class', y="Intensity", hue="class", ax=ax)
                    # sort so colors match between decision tree and boxplots
                    ax = sns.boxplot(data=unpacked_df, x='class', y="Intensity", hue="class", ax=ax, hue_order=np.unique(unpacked_df['class']))

                    pval_rows = corrected_pvals_by_pdmn[corrected_pvals_by_pdmn['peak_id'] == peak_id]
                    cor_pvals = []
                    reg_pvals = []
                    box_pairs = []
                    class_comparisons = pval_rows['class_comparison']
                    y = 0.035

                    for class_comparison in class_comparisons:
                        pval_row = pval_rows[pval_rows['class_comparison'] == class_comparison]

                        p_value = pval_row['raw_p_values'].values[0]
                        corrected_pval = pval_row['corrected_p_values'].values[0]

                        if len(class_comparisons) == 1:
                            reg_pvals.append(p_value)
                            cor_pvals.append(corrected_pval)
                            classes = class_comparison.split("_VS_")
                            box_pairs.append(((classes[0],classes[0]), (classes[1],classes[1])))
                        else:
                            # convert pvalue to scientific notation instead of <0.001
                            if p_value < 0.001:
                                plt.figtext(x=0.03, y=y,
                                            s=f"{class_comparison}: p = {BoxPlot.convert_pval(p_value)} | fdr_corrected = {BoxPlot.convert_pval(corrected_pval)}")
                            else:
                                plt.figtext(x=0.03, y=y, s="{}: p ={:.4f}  | fdr_corrected = {:.4f}".format(class_comparison, p_value, corrected_pval))
                            y -= 0.04

                    if len(class_comparisons) == 1:
                        # only works in binary case - one vs rest doesnt make sense in this annotation
                        statannot.add_stat_annotation(
                            ax, data=unpacked_df, x='class', y='Intensity', hue='class',
                            box_pairs=box_pairs, text_format='full',
                            perform_stat_test=False, pvalues=cor_pvals, loc='inside', verbose=0)

                        text_annot_custom = "P-Values are FDR corrected using the Benjamini-Hochberg method"
                        plt.figtext(x=0.03, y=y, s=text_annot_custom)

                    # remove x-axis label
                    ax.set_xlabel('')
                    # put legend outside plot
                    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
                    plt.title(peak_id)

                    if plot_parameters.get('use_buffer', False):
                        fig_name = f"{Path(plot_parameters['plot_prefix']).stem}_bp_{peak_id}.png"
                    else:
                        fig_name = f"{figure_dir}/{plot_parameters['plot_prefix']}_{peak_detection_name}_model_of_class_{model_of_class}_{peak_id}.png"

                    fig = plt.gcf()
                    plots_to_return.append(((PerformanceMeasure.FDR_CORRECTED_P_VALUE.name, peak_detection_name,
                                             model_of_class, peak_id), save_plot_to_buffer(plot_parameters, fig),
                                            fig_name))
                    if not plot_parameters.get('use_buffer', False):
                        fig.savefig(fig_name, dpi=300, bbox_inches='tight')
                    plt.close()

        return plots_to_return

    BoxPlotBestFeature = _plot_boxplot_of_best_features_seaborn


class TreePlot(object):
    """
    Create plot of tree with graphviz
    """
    @staticmethod
    def _plot_decision_trees_from_analysis_result(analysis_result, plot_parameters, limit_to_peak_detection_method_name=False):
        print(f"Plotting Decision trees buffer = {plot_parameters.get('use_buffer', False)} for {analysis_result}")
        # plot decision tree for each performance_measure and peak detection method
        # assert isinstance(analysis_result, AnalysisResult)
        if plot_parameters.get('use_buffer', False):
            tmp = os.path.join(tempfile.gettempdir(), '.breath/{}'.format(hash(os.times())))
            os.makedirs(tmp)
            plot_parameters['plot_dir'] = tmp

        # create temporary dir to save plots, read them back in, delete tempdir
        plots = []
        dot_data_full = analysis_result.decision_tree_per_evaluation_method_and_peak_detection
        for eval_method_name, by_pdm_dot_data in dot_data_full.items():
            for pdm_name, dot_data_lis in by_pdm_dot_data.items():

                # forgot the else case for non-automatic workflow - limit is not set
                if (limit_to_peak_detection_method_name and (limit_to_peak_detection_method_name == pdm_name)) or (not limit_to_peak_detection_method_name):
                    plots.append(TreePlot._plot_decision_tree_helper(dot_data_lis, eval_method_name=eval_method_name, pdm_name=pdm_name, plot_parameters=plot_parameters))

        if plot_parameters.get('use_buffer', False):
            shutil.rmtree(tmp, ignore_errors=True)
        return plots

    @staticmethod
    def _plot_decision_tree_helper(dot_data, eval_method_name, pdm_name, plot_parameters):
        figure_dir = f"{plot_parameters['plot_dir']}/decision_trees"
        fig_path = f"{figure_dir}/{plot_parameters['plot_prefix']}_{eval_method_name}_{pdm_name}"
        rv = []
        for class_comparison_str, dots in dot_data:
            if plot_parameters.get('use_buffer', False):
                # fig_name = f"{Path(plot_parameters['plot_prefix']).stem}_dt_{class_comparison_str}.png"
                # fig_name = f"{Path(plot_parameters['plot_prefix']).stem}_dt.png"
                # fig_name = f"{Path(plot_parameters['plot_prefix']).stem}_dt.png"
                fig_name = f"{Path(fig_path).stem}_dt.png"
            else:
                # fig_name = f"{fig_path}_{class_comparison_str}.png"
                fig_name = f"{fig_path}.png"
            rv.append(((eval_method_name, pdm_name, class_comparison_str), TreePlot._render_decision_tree_graph(dots, fig_name, plot_parameters), fig_name))
        return rv

    @staticmethod
    def _render_decision_tree_graph(dot_data, path, plot_parameters):
        rv = None
        graph = graphviz.Source(source=dot_data, format="png")
        graph.render(path[:-4], cleanup=True)

        # if use buffer - we write to temp dir and have to read them back in
        if plot_parameters.get('use_buffer', False):
            with open(path, 'rb') as bytes_handle:
                rv = BytesIO(bytes_handle.read())
        return rv

    DecisionTrees = _plot_decision_trees_from_analysis_result





def save_plot_to_buffer(plot_parameters, fig, dpi=200):
    rv = None
    if plot_parameters.get('use_buffer', False):
        buffer = BytesIO()
        fig.savefig(buffer, bbox_inches='tight', dpi=dpi)
        rv = buffer
    return rv



import glob
import os
from pathlib import Path

import numpy as np
import pandas as pd
import pyopenms as oms

from .ProcessingMethods import GCMSPeakDetectionMethod

"""
Nomenclature: 
    Spectrum: contains all points with a single RT value
    TIS: total ion spectrum, sum of intensity over all spectra
    XIC: extracted ion chromatogram - basically a subset of ms1 spectrum with smaller m/z range
    BPC: Base peak chromatogram - contains mmost intense signal for each RT across all m/z
    Isotope trace: signal that is produced by single ion of a single analute at a particvular charge state
    isotopic envelope trace: group of isotropic traces produced by a single analyte at a particular charge state
    Peak, feature, chromatogram extremely overloaded
"""
class UnsupportedGCMSFiletypeError(ValueError):
    """
    Error when not mzML or mzXML format used
    """


def load_ms_file(ms_experiment, path_to_file):
    """
    Load mzML / mzXML file into given ms_experiment object, filter to MSLevel = 1 - as only those supported for peakFinders
    :param ms_experiment:
    :param path_to_file:
    :param set_options:
    :return:
    """
    options = oms.PeakFileOptions()
    options.setMSLevels([1])  # MS1 is comparable to mcc-ims input
    # RuntimeError: FeatureFinder can only operate on MS level 1 data. Please do not use MS/MS data

    # support mzXML and mzML -file
    suffix = Path(path_to_file).suffix.lower()
    if suffix == ".mzxml":
        fh = oms.MzXMLFile()
    elif suffix == ".mzml":
        fh = oms.MzMLFile()
    else:
        raise UnsupportedGCMSFiletypeError("Unsupported filetype. Only mzXML and mzML format is supported.")

    fh.setOptions(options)
    # load data into experiment

    fh.load(str(path_to_file), ms_experiment)  # if problems loading - let pyopenms error bubble up
    ms_experiment.updateRanges()

    # get keys function loads the return into an empty list - very straightforward to guess
    # k = []
    # ms_experiment[0].getKeys(k)
    # ms_experiment[1].getKeys(k)
    # ms_experiment[2].getKeys(k)
    # ms_experiment[3].getKeys(k)
    # list_functions(ms_experiment[0])
    #
    # ms_experiment[0].getAcquisitionInfo()#
    #
    # oms.CachedmzML.store("myCache.mzML", ms_experiment)
    #
    # # Now load data
    # cfile = oms.CachedmzML()
    # oms.CachedmzML.load("myCache.mzML", cfile)
    #
    # meta_data = cfile.getMetaData()
    # meta_data.metaRegistry()
    # help(meta_data)
    # list_functions(meta_data)
    # meta_data.getKeys(k)
    #
    #
    # k = []
    # list_functions(\
    #     ms_experiment[0].getDataProcessing()[0].getMetaValue(k))
    # print(k)
    # list_functions(fh.getOptions())
    # try centroided approach to pickPeaks
    #   if not centroided,
    #       apply smoothing and peak detection to form centroided data = peakMap / FeatureXML file
    return ms_experiment


def detect_peaks_gcms_centroid(ms_experiment, parameters, debug=False):
    """
    Applicable to centroided experiments, also see https://abibuilder.informatik.uni-tuebingen.de/archive/openms/Documentation/nightly/html/a16103.html
    :param ms_experiment:
    :param parameters:
    :return:
    """
    print(f"Detecting peaks with {GCMSPeakDetectionMethod.CENTROIDED}")
    ff = oms.FeatureFinder()

    if not debug:
        ff.setLogType(oms.LogType.NONE)
    else:
        ff.setLogType(oms.LogType.CMD)

    # Run the feature finder
    name = "centroided"
    pdm_name = GCMSPeakDetectionMethod.CENTROIDED.name
    parameters['detection_mode'] = name
    parameters['pdm_name'] = pdm_name
    # name = parameters['detection_mode']
    features = oms.FeatureMap()
    seeds = oms.FeatureMap()
    ff_params = oms.FeatureFinder().getParameters(name)
    ff.run(name, ms_experiment, features, ff_params, seeds)
    # features.setUniqueIds()
    features.ensureUniqueId()
    fh = oms.FeatureXMLFile()
    feature_storage_path = f"{parameters['input_filename']}_output.featureXML"
    fh.store(feature_storage_path, features)
    parameters['feature_storage'] = feature_storage_path
    print("Found", features.size(), "features")
    return parameters


def detect_peaks_gcms_isotopewavelet(ms_experiment, parameters, debug=False):
    """
    Use isotop wavelet to process raw data - can perform poorly on centroided data
        - also see https://abibuilder.informatik.uni-tuebingen.de/archive/openms/Documentation/nightly/html/a16105.html
    TODO use "corrected" intensity_type
    :param ms_experiment:
    :param parameters:
    :return:
    """
    print(f"Detecting peaks with {GCMSPeakDetectionMethod.ISOTOPEWAVELET}")
    ff = oms.FeatureFinder()
    if not debug:
        ff.setLogType(oms.LogType.NONE)
    else:
        ff.setLogType(oms.LogType.CMD)

    # Run the feature finder
    name = "isotope_wavelet"
    pdm_name = GCMSPeakDetectionMethod.ISOTOPEWAVELET.name
    parameters['detection_mode'] = name
    parameters['pdm_name'] = pdm_name
    # name = parameters['detection_mode']
    features = oms.FeatureMap()
    seeds = oms.FeatureMap()
    ff_params = ff.getParameters(name)

    # complains about "the extremal length of the wavelet is larger (47661) than the number of data points"
    # wavelet_length is defined by mz_cutoff / min_spacing
    # hr_data must be true if high-resolution data (orbitrap, FTICR)
    # hr_data parameter for isotopewavelet function
    is_hr_data = parameters.get("hr_data", False)
    if is_hr_data:
        hr_key = b"hr_data"  # hr_data takes extremely long - >= 2h per measurement of (!)32MB - there are way larger spectra...
        ff_params.setValue(hr_key, b"true")

    ff.run(name, ms_experiment, features, ff_params, seeds)
    features.setUniqueIds()
    fh = oms.FeatureXMLFile()
    feature_storage_path = f"{parameters['input_filename']}_output.featureXML"
    fh.store(feature_storage_path, features)
    parameters['feature_storage'] = feature_storage_path
    print("Found", features.size(), "features")
    return parameters


# two peak pickers - PeakPickerWavelet and PeakPickerHiRes
def detect_peaks_gcms_peak_picker_wavelet(ms_experiment, parameters):
    """
    Use isotop wavelet to process raw data - can perform poorly on centroided data
        https://abibuilder.informatik.uni-tuebingen.de/archive/openms/Documentation/nightly/html/a16159.html
    TODO use "corrected" intensity_type
    :param ms_experiment:
    :param parameters:
    :return:
    """
    #set estimate_peak_width to true


    # Run the feature finder
    name = "peak_picker_wavelet"
    parameters['detection_mode'] = name


    # outdated code in https://github.com/OpenMS/pyopenms-extra/blob/master/src/examples/peakpicker_scipyFFT.py

    # pick spectrum
    ff = oms.FeatureFinder()
    ff.setLogType(oms.LogType.NONE)

    algo_name = oms.FeatureFinderAlgorithmIsotopeWavelet().getProductName()
    # picker = oms.FeatureFinderAlgorithmIsotopeWavelet()

    feature_map = oms.FeatureMap()
    seeds = oms.FeatureMap()

    # seeds = FeatureMap()
    algo_params = ff.getParameters(algo_name)
    ff.run(algo_name, ms_experiment, feature_map, algo_params, seeds)
    feature_map.setUniqueIds()

    parameters = default_store_feature_xml(feature_map, parameters)

    return parameters


def filter_feature_xmls(dir, name_list=[], include_aligned=0):
    """
    Returns list of feature xmls that are not aligned yet , so ending with "ML_output.featureXML"
    :param dir:
    :return:
    """
    if include_aligned:
        suffix_to_check = get_default_feature_xml_storage_suffix()  # includes aligned suffixes
    else:
        suffix_to_check = "ML_output.featureXML"  # only unaligned featureXMLS

    if name_list:
        matching_feature_xmls = [fn for fn in name_list if str(fn).endswith(suffix_to_check)]
    else:
        matching_feature_xmls = glob.glob(str(dir) + f"/*{suffix_to_check}")
    return sorted(matching_feature_xmls)


# needs to match class_labels with the feature_xml_lis - otherwise problem with feature_matrix index
def align_feature_xmls(feature_xml_lis, consensus_map_out_path="", class_label_dict={}):
    """
    first apply pose clustering to include all features maps
      next link/group them across all features

    Each MS1 spectrum from raw-file will create a feature file -
    we need to load and align them to get unique and representative features
    :param feature_xml_lis:
    :param consensus_map_out_path:
    :return: consensus_map, consensus_map_out_path, measurement_names
    """
    # do consensus map normalization and export -
    # can't hack normalization together from lack of example usage and poor signature
    #   - no normalization implemented

    # openms won't deal with posix paths - wants to have strings instead
    # need to make sure it get's those
    # let's sort them to make sure feature matrix is also sorted
    feature_xml_lis = sorted([str(fx) for fx in feature_xml_lis])

    num_features_list = []
    for current_feature_xml_path in feature_xml_lis:
        # load features into FeatureMaps
        cm = oms.FeatureMap()  # current_map
        oms.FeatureXMLFile().load(current_feature_xml_path, cm)
        # list_functions(current_map, prefix="")
        num_features_list.append(cm.size())
        del cm

    # should choose the feature file / experiment with most features as reference
    max_index = np.argmax(num_features_list)
    reference_map_path = feature_xml_lis[max_index]

    default_max_num_peaks_considered = 1000
    default_max_scaling_value = 10.0
    aligned_paths = []
    for current_feature_xml_path in feature_xml_lis:
        # load features into FeatureMaps
        reference_map = oms.FeatureMap()  # pairwise alignment - so need master map -
        oms.FeatureXMLFile().load(reference_map_path, reference_map)

        current_map = oms.FeatureMap()
        oms.FeatureXMLFile().load(current_feature_xml_path, current_map)

        # create a transformation description required as init for aligner
        transformation_description = oms.TransformationDescription()

        # adjust max scaling parameter otherwise leads to error when running with algae samples
        # adjust max num peaks to 2k - also would leads to error when running with algae samples

        aligner = oms.MapAlignmentAlgorithmPoseClustering()
        aligner_params = aligner.getParameters()

        # print(aligner_params.asDict().keys())
        max_scaling_key = b'superimposer:max_scaling'
        # aligner_params.getEntry(max_scaling_key)
        aligner_params.setValue(max_scaling_key, default_max_scaling_value)

        max_num_peaks_key = b'max_num_peaks_considered'
        # aligner_params.getEntry(max_num_peaks_key)
        aligner_params.setValue(max_num_peaks_key, default_max_num_peaks_considered)  # default = 1000
        # need higher default for algae

        # decrease runtime by removing weak signals
        # print(aligner_params.asDict())
        num_used_points_key = b'superimposer:num_used_points'
        # aligner_params.getEntry(num_used_points_key)
        aligner_params.setValue(num_used_points_key, 1000)  # half the default parameter, speed up alignment

        aligner.setParameters(aligner_params)

        aligner.setReference(reference_map)

        try:
            # run alignment
            aligner.align(current_map, transformation_description)
        except RuntimeError as re:
            if 'max_num_peaks_considered' in str(re):
                # retry with higher threshold - probably algae dataset
                default_max_num_peaks_considered = 15000  # 15 fold - makes it a lot slower but less error prone
                aligner_params.setValue(max_num_peaks_key, default_max_num_peaks_considered)
                default_max_scaling_value = 20.0  # need to increase to 20
                aligner_params.setValue(max_scaling_key, default_max_scaling_value)
                print(f"Encountered GC/MS Clustering issue - setting 'max_num_peaks_considered' to {default_max_num_peaks_considered} and 'superimposer:max_scaling' to {default_max_scaling_value}")
                aligner.setParameters(aligner_params)
                aligner.setReference(reference_map)
                aligner.align(current_map, transformation_description)

        current_map.updateRanges()
        reference_map.updateRanges()

        # update feature XML files - both reference and current
        updated_current_map_path = default_store_aligned_feature_xml(current_map, current_feature_xml_path)
        updated_reference_path = default_store_aligned_feature_xml(reference_map, reference_map_path)
        reference_map_path = updated_reference_path

        aligned_paths.append(updated_current_map_path)

    # also replace here with new reference we updated the reference map to
    aligned_paths[max_index] = reference_map_path

    #   link/group them across features to create consensus map

    grouper = oms.FeatureGroupingAlgorithmUnlabeled()
    # leave parameters default

    # according to openms documentation:
    #   b) Call "setReference", "addToGroup" (n times), "getResultMap" in that order.

    for i, current_feature_map_path in enumerate(aligned_paths):
        current_map = oms.FeatureMap()
        oms.FeatureXMLFile().load(current_feature_map_path, current_map)

        if not i:
            # first iteration - use as reference
            grouper.setReference(i, current_map)

        else:
            grouper.addToGroup(i, current_map)

    # get consensus map
    consensus_map = grouper.getResultMap()

    # consensus map requires some mapping between ids and filenames - otherwise will complain
    print(f"Mapping aligned results back to class labels")
    class_label_fns = list(class_label_dict.keys())
    fds = {i: oms.ColumnHeader() for i, _ in enumerate(aligned_paths)}
    measurement_names = []
    for i, aligned_path in enumerate(aligned_paths):
        # fds[i].filename = b"file0"
        current_fn = f"{str(Path(aligned_path).stem)}{str(Path(aligned_path).suffix)}"

        # this is where we need to replace the feature_xml filenames with the ones from class_labels
        if class_label_dict:
            # could do longest substring match with each of the fns in class_label dict to find matching filename
            #   django will rename duplicate filenames instead of overwriting
            # or we expect both featureXML input and class_label_dict to be ordered - which they should be when using the getter
            fds[i].filename = class_label_fns[i]

        else:
            fds[i].filename = current_fn.encode("UTF8")  # needs bytestring representation

        measurement_names.append(current_fn)

    consensus_map.setColumnHeaders(fds)

    #  cleanup aligned_feature_xmls - can be >30mb per file - so better remove them
    for ap in aligned_paths:
        os.remove(ap)

    #   do consensus map normalization and export to consensus files
    # using median normalization, also available are Quantile and "robust regression"
    normalizer = oms.ConsensusMapNormalizerAlgorithmMedian()

    # ConsensusMapNormalizerAlgorithmMedian
    # signature of class is more than incomplete ... *args **kwargs for required parameters is not the best implementation choice...
    # but gives TypeError requiring int when calling with
    # normalizer.normalizeMaps(consensus_map, "NM_SCALE", "", "") #
    """
    normalizer.normalizeMaps(map, method, acc_filter, desc_filter)
    map	ConsensusMap
    method	whether to use scaling or shifting to same median 
    acc_filter	string describing the regular expression for filtering accessions
    desc_filter	string describing the regular expression for filtering descriptions 
    """
    """
        method: probably 0 / 1 - referenced as Enumerator in OpenMS documentation
        from shell output can deduce normalization methods are
        0: NM_SCALE 	scale to same median using division/multiplication  
        1: NM_SHIFT 	shift using subtraction/addition
    """
    normalizer.normalizeMaps(consensus_map, 0, "", "")

    # don't export if not required - requires more file management
    # now export
    if consensus_map_out_path:
        oms.ConsensusXMLFile().store(str(consensus_map_out_path), consensus_map)

    return consensus_map, measurement_names


def convert_consensus_map_to_feature_matrix(consensus_map, measurement_names, rounding_precision=0):
    """
    Create a `pandas.DataFrame` from `pyopenms.ConsensusMap`. Col-index holds PeakIds and row-index measurement names
        without suffix
    :param consensus_map: `pyopenms.ConsensusMap`
    :param measurement_names: measurement names in same order as referenced in consensus map index
    :return:
    """
    all_feature_records = []
    peak_ids = []
    for consensus_feature in consensus_map:
        # consensus_feature.computeConsensus()
        peak_id = generate_consensus_peak_id(consensus_feature, rounding_precision=rounding_precision)
        # print("ConsensusFeature", peak_id)

        # consensus feature clusters are retrieved - but don't have unique or persistent ids - concat f"{rt}_{mz}_{charge}" as feature id
        # consensus_map.getColumnHeaders()[0]

        single_col_dict = {}
        # now get intensity for each of the files
        # lets just assume that they are always in order
        for single_feature in consensus_feature.getFeatureList():
            # print(" -- Feature", single_feature.getMapIndex(), round(single_feature.getIntensity(), 3))
            single_col_dict[single_feature.getMapIndex()] = round(single_feature.getIntensity(), 3)

        all_feature_records.append((peak_id, single_col_dict))
        peak_ids.append(peak_id)

    # add all to pandas dataframe
    # sort by peak_id
    sorted_feature_records = sorted(all_feature_records, key=lambda t: t[0])
    num_rows = len(measurement_names)
    feature_df = pd.DataFrame()
    for (peak_id, col_dict) in sorted_feature_records:
        numpy_col = np.zeros(num_rows, dtype=float)
        for index, intensity in col_dict.items():
            numpy_col[index] = intensity
        feature_df[peak_id] = numpy_col

    # index_names = [mn.split(".mz")[0] + ".mzML" for mn in measurement_names]
    # shouldn't assume only mzml - cn also be mzxml!
    # index_names = [str(mn).split(".mz")[0] for mn in measurement_names]
    # dont split - use measurement names passed
    index_names = measurement_names
    feature_df.index = index_names
    return feature_df


def default_store_feature_xml(feature_map, parameters):
    """
    Store feature xml file in fedault location with default name - standaradized manner
    :param feature_map:
    :param parameters:
    :return: updated parameters with 'feature_storage' set to path
    """
    fh = oms.FeatureXMLFile()
    feature_xml_outname = f"{parameters['input_filename']}{get_default_feature_xml_storage_suffix()}"
    fh.store(feature_xml_outname, feature_map)
    parameters['feature_storage'] = feature_xml_outname
    return parameters

def default_store_aligned_feature_xml(feature_map, original_path):
    """
    Store feature xml file in default location with default aligned name
    :param feature_map:
    :param parameters:
    :return: updated parameters with 'feature_storage' set to path
    """
    fh = oms.FeatureXMLFile()
    storage_suffix = get_default_feature_xml_storage_suffix('aligned')
    if str(original_path).endswith(storage_suffix):
        # overwrite - probably writing the reference feature map
        feature_xml_outname = original_path
    else:
        feature_xml_outname = f"{Path(original_path).parent}/{Path(original_path).stem}{storage_suffix}"

    fh.store(feature_xml_outname, feature_map)
    return feature_xml_outname


def get_default_feature_xml_storage_suffix(suffix_prefix=""):
    """"
    :return: Default storage suffix for feature xml files
    """
    if suffix_prefix:
        return f"_{suffix_prefix}_output.featureXML"
    else:
        return f"_output.featureXML"

def list_functions(obj, prefix="get"):
    for fun in dir(obj):
        if prefix:
            if str(fun).startswith(prefix):
                print(fun)
        else:
            if not str(fun).startswith("__"):
                print(fun)


def denoise_sgf_gcms(ms_experiment):
    # chroms = ms_experiment
    sg = oms.SavitzkyGolayFilter()
    sg.filterExperiment(ms_experiment)
    # out_path = f"{path_to_mzml}.filter.mzML"
    # # TODO they changed the order of arguments in the store method in v2.5 - be aware!
    # oms.MzMLFile().store(out_path, chroms)
    return ms_experiment


def generate_consensus_peak_id(consensus_feature, rounding_precision=0):
    """
    Generate the consonsus peak id by concatenating rt_mz_charge - trailing zeros for sortability to reproduce same ordering
    :param consensusPeak:
    :return:
    """
    rp = rounding_precision
    if rp:
        return f"Peak_{round(consensus_feature.getRT(),rp):05}_{round(consensus_feature.getMZ(),rp):05}_{int(consensus_feature.getCharge())}"
    else:  # remove the .0 from the peak id - as we're rounding the digits
        rt = int(round(consensus_feature.getRT(), rp))
        mz = int(round(consensus_feature.getMZ(), rp))
        ch = int(consensus_feature.getCharge())
        return f"Peak_{rt}_{mz}_{ch}"


def filter_mzml_or_mzxml_filenames(dir, filelis=[]):
    """
    Return all mz and mzxml filenames, if filelis supplied (eg from namelist of zip - return suset of them
    :param dir:
    :param filelis:
    :return:
    """
    if not filelis:
        if not str(dir).endswith("/"):
            dir = str(dir) + "/"

        paths_mzml = glob.glob(dir + "*.mzML")
        paths_mzml2 = glob.glob(dir + "*.mzml")

        paths_mzxml = glob.glob(dir + "*.mzXML")
        paths_mzxml2 = glob.glob(dir + "*.mzxml")

        all_paths = []
        all_paths.extend(paths_mzml)
        all_paths.extend(paths_mzml2)
        all_paths.extend(paths_mzxml)
        all_paths.extend(paths_mzxml2)

    else:
        all_paths = []
        for fn in filelis:
            if str(fn).lower().endswith(".mzml") or str(fn).lower().endswith(".mzxml"):
                all_paths.append(fn)

    return all_paths


def detect_raw_peaks_wavelet_single_core(fn_in):
    """
    Apply wavelet transformation and pick peaks on raw MS1 spectra, expecting
    :param parameters:
    :return: filename of feature xml file
    """
    current_params = dict()
    current_params['input_filename'] = fn_in
    fn_out = run_raw_wavelet_sample(current_params)['feature_storage']
    print(f"Saved features to {fn_out}")
    return fn_out


def run_centroid_sample(parameters):
    """
    Run centroid gcms detection. Expects 'input_filename' to be set. Use 'feature_storage' to access result fn
    :returns dict('feature_storage': 'fn')
    :param parameters:
    :return: dict()
    """
    ms_experiment = oms.MSExperiment()
    ms_experiment = load_ms_file(ms_experiment=ms_experiment, path_to_file=parameters['input_filename'])
    return detect_peaks_gcms_centroid(ms_experiment=ms_experiment, parameters=parameters)


def run_raw_wavelet_sample(parameters):
    """
    Run wavelet based gcms detection. Expects 'input_filename' to be set. Use 'feature_storage' to access result fn
    :returns dict('feature_storage': 'fn')
    :param parameters:
    :return: dict()
    """
    ms_experiment = oms.MSExperiment()
    ms_experiment = load_ms_file(ms_experiment=ms_experiment, path_to_file=parameters['input_filename'])
    # ms_experiment = denoise_sgf_gcms(ms_experiment)
    return detect_peaks_gcms_isotopewavelet(ms_experiment=ms_experiment, parameters=parameters)



from abc import ABCMeta
from enum import Enum, auto
### MCCIMS PREPROCESSING METHODS

class MethodSelection(Enum):
    """
    Base Method Selection
    """
    __metaclass__ = ABCMeta

    def _generate_next_value_(name, start, count, last_values):
        return name

    def __repr__(self):
        return self.name

    # if implementing __eq__ we also need to implement __hash__ otherwise it cannot be use in dicts andymore
    # https://stackoverflow.com/questions/1608842/types-that-define-eq-are-unhashable
    def __eq__(self, other):
        if isinstance(other, Enum):
            return self.value == other.value
        if isinstance(other, str):
            return str(self) == other
        else:
            return object.__eq__(Enum, other)

    def __hash__(self):
        return hash(self._name_)

class PreprocessingMethod(MethodSelection):
    """
    Base Preprocessing method
    """

class MccImsPreprocessingMethod(PreprocessingMethod):
    """
    Could be a PeakDetectionMethod, NormalizationMethod, DenoisingMethod or PeakAlignmentMethod
    """

class MeasurementAlignmentScalingMethod(MccImsPreprocessingMethod):
    """
    Scale measurements on retention time and inverse reduced mobility axis
    Used to reduce variation between instruments, laboratories and operation parameters
    Will be performed in Analysis
    """
    IRM_SCALING = auto()
    # additional method could be irm-alignment


class PeakDetectionMethod(MccImsPreprocessingMethod):
    """
    Available peak detection Methods for MCC Analysis.
    Will be performed in Analysis
    """
    WATERSHED = auto()
    TOPHAT = auto()
    JIBB = auto()
    VISUALNOWLAYER = auto()

    # look at possible methods at https://stackoverflow.com/questions/1713335/peak-finding-algorithm-for-python-scipy
    # eg scipy.signal.find_peaks_cwt
    # or https://gist.github.com/endolith/250860


class ExternalPeakDetectionMethod(MccImsPreprocessingMethod):
    """
    External peak detection Methods for MCC Analysis.
    Depend on third party libraries and external binaries - eg PEAX
    Will be performed in Analysis
    """
    PEAX = auto()
    CUSTOM = auto()
    # PEAX does it's own normalization, need to write Normalization parameters to parameters.cfg in directory from where peax is executed
    # DeepPeak


class DenoisingMethod(MccImsPreprocessingMethod):
    """
    Method to remove Noise from an MccImsMeasurement.
    Will be performed in Measurement
    """
    NOISE_SUBTRACTION = auto()
    MEDIAN_FILTER = auto()
    SAVITZKY_GOLAY_FILTER = auto()
    GAUSSIAN_FILTER = auto()
    CROP_INVERSE_REDUCED_MOBILITY = auto()
    DISCRETE_WAVELET_TRANSFORMATION = auto()



class NormalizationMethod(MccImsPreprocessingMethod):
    """
    Available normalization Methods for MCC Analysis
    Will be performed in Measurement
    """
    BASELINE_CORRECTION = auto()
    INTENSITY_NORMALIZATION = auto()


class PeakAlignmentMethod(MccImsPreprocessingMethod):
    """
    Available PeakAlignment Methods for MCC Analysis
    Mainly Clustering methods
    Will be performed in Analysis
    """
    WARD_CLUSTERING = auto()
    K_MEANS_CLUSTERING = auto()
    AFFINITY_PROPAGATION_CLUSTERING = auto()
    WINDOW_FRAME = auto()
    MEAN_SHIFT_CLUSTERING = auto()
    DB_SCAN_CLUSTERING = auto()
    PROBE_CLUSTERING = auto()
    # add Voronoi diagram of known substances

class PerformanceMeasure(MethodSelection):
    """
    Available PerformanceMeasure Methods for MCC-IMS / GCMS Analysis
    Will be performed in Analysis
    """
    RANDOM_FOREST_CLASSIFICATION = auto()  # with ROC and Decision tree of best features
    PCA_LDA_CLASSIFICATION = auto()  # with ROC and Decision tree of best features
    FDR_CORRECTED_P_VALUE = auto()  # Boxplots with p_values
    DECISION_TREE_TRAINING = auto()  # only affects decision tree, not classification of samples

class FeatureReductionMethod(MethodSelection):
    """
    FeatureReduction Methods
    """
    REMOVE_PERCENTAGE_FEATURES = auto()  # remove features that are not present in a certain percentage of measurements
    # REMOVE_CONTROL_PEAKS_POSITIVE = auto()  #  option to only keep features that are more intense in comp to control samples
    # REMOVE_CONTROL_PEAKS_NEGATIVE = auto()  #  option to only keep features that are less intense in comp to control samples
    # REMOVE_PRE_RIP_PEAKS = auto()  # option to only keep features that have irm >= 0.485


### GCMS PREPROCESSING METHODS

class GCMSPreprocessingMethod(PreprocessingMethod):
    """
    Preprocessing Method applicable for GCMS data - TODO include LCMS in class name?
    """

class GCMSPeakDetectionMethod(GCMSPreprocessingMethod):
    """
    Peak detection method for GCMS data - extract the features from profile or centroided data - also see https://abibuilder.informatik.uni-tuebingen.de/archive/openms/Documentation/nightly/html/a16047.html
    """
    CENTROIDED = auto() # FeatureFinderCentroided from pyopenms
    ISOTOPEWAVELET = auto() # FeatureFinderIsotopeWavelet from pyopenms -  "is only able to handle MS1 data"
    CUSTOM = auto()
    # TODO implement smoothing function, as isotopewavelet is slow and overly sensitive
    # TODO MRM? MULTIPLEX?
    # TODO FeatureFinderAlgorithmSH = SuperHirn
    #       centroids data, then does feature finding and feature merging
    #       extracts 1D peaks though

class GCMSAlignmentMethod(GCMSPreprocessingMethod):
    """
    Peak alignement method for GCMS data - align featureXML files into a consensus map
    """
    # TODO add different options
    POSE_FEATURE_GROUPER = auto()  # align features on axis with pose clustering, then use feature grouper to cluster - peak_ids are generated based on 2 decimals in rt, mz and


class GCMSSupportedDatatype(MethodSelection):
    """
    Supported datatypes Methods
    """
    RAW_MZML_OR_MZXML = auto()
    CENTROIDED_MZML_OR_MZXML = auto()



from collections import OrderedDict
import glob
from multiprocessing import Pool
import numpy as np
import os
from pathlib import Path
import pyopenms as oms
from shutil import rmtree
import tempfile

import joblib

from .BreathCore import MccImsAnalysis, GCMSAnalysis, PredictionModel, construct_default_processing_evaluation_steps, construct_default_parameters

from .GCMSTools import (
    align_feature_xmls,
    convert_consensus_map_to_feature_matrix,
    GCMSPeakDetectionMethod,
    list_functions,
    filter_mzml_or_mzxml_filenames,
    run_centroid_sample,
    filter_feature_xmls,
)
from .ProcessingMethods import GCMSAlignmentMethod
from ..view.BreathVisualizations import HeatmapPlot, ClusterPlot, MaskPlot, RocCurvePlot, BoxPlot, TreePlot


def run_alignment_test(feature_xml_lis):
    """
    Uses all FeatureXML to create peak alignment result
    :param parameters:
    :return:
    """
    # pass intermediary consensusXML filename
    # TODO do we need to store a real copy of the consensus xml and intermediaries - clean up after?
    consensus_outfile_name = f"{Path(feature_xml_lis[0]).parent}/test.consensusXML"
    consensus_map, measurement_names = align_feature_xmls(feature_xml_lis, consensus_outfile_name)
    feature_df = convert_consensus_map_to_feature_matrix(consensus_map, measurement_names)
    return feature_df


def run_algae_platform():
    """
    Run algae example from start to finish using BreathCore implementations and classes
    :return:
    """
    data_dir = "/home/philipp/dev/breathpy/data"
    algae_dir = f"{data_dir}/algae/"
    algae_files, class_label_dict = prepare_algae_sample_filenames(algae_dir, ['dark', 'replete', 'light'])
    # gcms_a = GCMSAnalysis(measurements=algae_files, preprocessing_steps=[GCMSPeakDetectionMethod.ISOTOPEWAVELET], dataset_name="algae_limited", dir_level=".", preprocessing_parameters={GCMSPeakDetectionMethod.ISOTOPEWAVELET: {"hr_data": True}}, performance_measure_parameters={})
    gcms_a = GCMSAnalysis(measurements=algae_files, preprocessing_steps=[GCMSPeakDetectionMethod.ISOTOPEWAVELET], dataset_name="algae", dir_level=".", preprocessing_parameters={GCMSPeakDetectionMethod.ISOTOPEWAVELET: {"hr_data": True}}, performance_measure_parameters={})
    gcms_a.set_class_label_dict(class_label_dict)
    gcms_a.detect_peaks_parallel(num_cores=4)
    # gcms_a.detect_peaks()
    # feature alignment fails - some nan error during pose map alignment
    feature_xml_lis = filter_feature_xmls(algae_dir)
    consensus_map, measurement_names = align_feature_xmls(feature_xml_lis, "test.out")

    # TODO use restore feature - dont want to keep rerunning peak detection to get to feature df
    # feature_df = gcms_a.align_peaks(GCMSAlignmentMethod.POSE_FEATURE_GROUPER)
    # print(feature_df)

def run_tiny_samples_parallel(num_cores=4, sample_dir="/home/philipp/Desktop/gcms_data/sample"):
    sample_dir = Path(sample_dir)

    # mini_sample_path_to_mzml = Path(str(urine_dir) + "/" + "tiny.pwiz.1.1.mzML")
    sample_path_to_mzml = sample_dir/"FeatureFinderCentroided_1_input.mzML"

    # use same tiny file several times
    example_fns = []
    example_params = []
    for i in range(3):
        current_sample_path = Path(f"{sample_dir}/{sample_path_to_mzml.stem}_{i}.mzML")
        example_fns.append(current_sample_path)
        example_params.append({'input_filename': current_sample_path})

    with Pool(num_cores) as p:
        feature_xml_dicts = p.map(run_centroid_sample, example_params)

    feature_xml_lis = []
    # list will not be the same order as before - so need to sort before aligning
    for params in feature_xml_dicts:
        feature_xml_fn = params['feature_storage']
        feature_xml_lis.append(feature_xml_fn)
    consensus_map, measurement_names = align_feature_xmls(feature_xml_lis, "test.out")
    feature_df = convert_consensus_map_to_feature_matrix(consensus_map, example_fns)
    print(feature_df)
    return feature_df


def prepare_algae_sample_filenames(algae_dir, limit_to_classes=["n-limited"]):
    """
    Return paths to algae files
    :param algae_dir:
    :return:
    """
    algae_files = filter_mzml_or_mzxml_filenames(algae_dir)

    # get class label dict from zip / archive

    label_dict_path = MccImsAnalysis.guess_class_label_extension(algae_dir)
    label_dict = MccImsAnalysis.parse_class_labels(label_dict_path)

    if not limit_to_classes:
        return algae_files, label_dict

    # apply filtering by class label
    else:
        filtered_label_dict = OrderedDict()
        sorted_algae_files = sorted(algae_files)
        # now also need to filter sorted algae files


        for fn, label in label_dict.items():
            if label in limit_to_classes:
                filtered_label_dict[fn] = label

        filtered_sorted_algae_files = []
        for fn in sorted_algae_files:
            if str(Path(fn).stem) + str(Path(fn).suffix) in filtered_label_dict:
                filtered_sorted_algae_files.append(fn)

        return filtered_sorted_algae_files, label_dict

# use https://pyopenms.readthedocs.io/en/latest/datastructures_quant.html
#   need to differentiate between centroided and raw data - if centroided cant apply savitzky golay
#  user needs to tell us whether centroided data or not -
#  is encoded in mzML-file file description tag -
#  (cvParam cvRef="MS" accession="MS:1000127" name="centroid spectrum") -
#   - but can't find documentation to access cvParams to check - there is CVReference and CVMapping class but how to access / initialize these...


def run_gcms_platform_multicore(sample_dir, preprocessing_params={}, evaluation_parms={}, num_cores=4):
    """
    Run platform with default parameters unless specified -
        will look for existing results in results/data/sample_dir and if sample_dir starts with 'train_' if will use the
        same path with 'test_' for prediction
    :param sample_dir:
    :return:
    """
    dataset_name = str(Path(sample_dir).stem)
    print(f"Starting analysis of {dataset_name}")
    plot_params, file_params = construct_default_parameters(dataset_name, folder_name=dataset_name, make_plots=True)

    raw_files, class_label_dict = GCMSAnalysis.prepare_raw_filenames_class_labels(sample_dir)
    gcms_a = GCMSAnalysis(measurements=raw_files,
                          preprocessing_steps=preprocessing_params.keys(),
                          dataset_name=dataset_name, dir_level=".",
                          preprocessing_parameters=preprocessing_params,
                          performance_measure_parameters=evaluation_parms)
    gcms_a.set_class_label_dict(class_label_dict)

    feature_xml_lis = filter_feature_xmls(sample_dir)
    if not feature_xml_lis:
        gcms_a.detect_peaks_parallel(num_cores=num_cores)
        # gcms_a.detect_peaks()
    feature_xml_lis = filter_feature_xmls(sample_dir)

    if len(class_label_dict) != len(feature_xml_lis):
        raise ValueError(f"Class labels don't match feature XMLs {len(class_label_dict)} != {len(feature_xml_lis)}.")
    # print(feature_xml_lis)

    # cross validation and plots
    # import peak detection results - but don't use aligned ones if already present - will lead to problems downstream
    gcms_a.import_peak_detection_results(feature_xml_lis)

    # Next: create Feature matrix
    # import feature matrix if already existent - otherwise do alignment
    result_data_dir = f"{gcms_a.dir_level}/results/data/{gcms_a.dataset_name}/"
    try:
        # can take several minutes to load - some alignment dfs are hundred of MBs in size
        fm = gcms_a.import_alignment_result(result_data_dir, gcms_a.peak_detection_steps[0].name)
    except FileNotFoundError:
        fm = gcms_a.align_peaks(GCMSAlignmentMethod.POSE_FEATURE_GROUPER)
    # consensus_map, measurement_names = align_feature_xmls(feature_xml_lis, f"{dataset_name}.out", class_label_dict)

    ims_analysis = gcms_a.prepare_custom_fm_approach(fm)

    # expects a FloatPeakAlignmentResult - which we never have coming from GCMS
    # ims_analysis.reduce_features(ims_analysis.AVAILABLE_FEATURE_REDUCTION_METHODS)
    ims_analysis.remove_redundant_features_fm()
    # reduced_fm_dict = ims_analysis.analysis_result.remove_redundant_features_fm(ims_analysis.analysis_result)
    ims_analysis.evaluate_performance(exising_analysis_result=ims_analysis.analysis_result)
    # if len(set(ims_analysis.analysis_result.class_labels)) == 2:
    #     RocCurvePlot.ROCCurve(ims_analysis.analysis_result, plot_parameters=plot_params)
    # else:
    #     RocCurvePlot.MultiClassROCCurve(ims_analysis.analysis_result, plot_parameters=plot_params)
    # BoxPlot.BoxPlotBestFeature(ims_analysis.analysis_result, plot_parameters=plot_params)
    if plot_params['make_plots']:
        TreePlot.DecisionTrees(ims_analysis.analysis_result, plot_parameters=plot_params)

    # continue with prediction and preprocessing of test set
    tmp = os.path.join(tempfile.gettempdir(), '.breath/{}'.format(hash(os.times())))
    os.makedirs(tmp)

    dataset_name = str(Path(sample_dir).stem)
    if dataset_name.startswith("train_"):
        # actually using sample_dir which is at a different location
        test_dir = str(sample_dir).replace("train_", "test_")
        test_dataset_name = dataset_name.replace("train_", "test_")
        test_result_data_dir = f"{gcms_a.dir_level}/results/data/{gcms_a.dataset_name}/".replace("train_", "test_")
    else:
        test_dir = sample_dir
        test_dataset_name = dataset_name
        test_result_data_dir = f"{gcms_a.dir_level}/results/data/{gcms_a.dataset_name}/"

    #  handle for GCMS files
    predictor_path = tmp + "/pred_model.sav"  # export and reload for usage later on
    ims_analysis.analysis_result.export_prediction_models(path_to_save=predictor_path)
    predictors = joblib.load(predictor_path)

    # clean up tempdir
    rmtree(tmp, ignore_errors=True)

    test_files, test_class_label_dict = GCMSAnalysis.prepare_raw_filenames_class_labels(test_dir)

    test_gcms_a = GCMSAnalysis(measurements=test_files,
                          preprocessing_steps=preprocessing_params.keys(),
                          dataset_name=test_dataset_name, dir_level=".",
                          preprocessing_parameters=preprocessing_params,
                          performance_measure_parameters=evaluation_parms)
    test_gcms_a.set_class_label_dict(test_class_label_dict)

    # convert to feature matrix - or load if available - fallback to detection import + alignment, fallback to full pipeline

    test_feature_xml_lis = filter_feature_xmls(test_dir)
    if not test_feature_xml_lis:
        gcms_a.detect_peaks_parallel(num_cores=4)
    test_feature_xml_lis = filter_feature_xmls(test_dir)

    if len(test_class_label_dict) != len(test_feature_xml_lis):
        raise ValueError(f"Class labels don't match feature XMLs {len(test_class_label_dict)} != {len(test_feature_xml_lis)}.")
    print(test_feature_xml_lis)

    # import peak detection results - but don't use aligned ones if already present - will lead to problems downstream
    gcms_a.import_peak_detection_results(test_feature_xml_lis)

    # Next: create Feature matrix

    # import feature matrix if already existent - otherwise do alignment
    try:
        # can take several minutes to load - some alignment dfs are hundred of MBs in size
        test_fm = gcms_a.import_alignment_result(test_result_data_dir, gcms_a.peak_detection_steps[0].name)

    except FileNotFoundError:
        test_fm = gcms_a.align_peaks(GCMSAlignmentMethod.POSE_FEATURE_GROUPER)
    # consensus_map, measurement_names = align_feature_xmls(feature_xml_lis, f"{dataset_name}.out", class_label_dict)

    test_ims_analysis = gcms_a.prepare_custom_fm_approach(test_fm)

    # make sure features match
    train_column_names = ims_analysis.analysis_result.trainings_matrix[gcms_a.peak_detection_steps[0].name].columns

    trainings_feature_names_by_pdmn = {gcms_a.peak_detection_steps[0].name: train_column_names}

    # make test_fm match fm
    # test_fm_dict = {gcms_a.peak_detection_steps[0].name: test_fm}
    test_fm_dict = test_ims_analysis.analysis_result.feature_matrix
    df_dict_matching_training = PredictionModel.reconstruct_remove_features(test_fm_dict, trainings_feature_names_by_pdmn)

    predictionModel = PredictionModel(
        preprocessing_params=preprocessing_params,
        evaluation_params=ims_analysis.performance_measure_parameter_dict,
        scipy_predictor_by_pdm=predictors,
        feature_names_by_pdm=ims_analysis.analysis_result.feature_names_by_pdm,)  # not sure whether important to use train / test matrix

    prediction, test_matrix_by_pdm = predictionModel.predict_from_feature_matrix(pdms=[GCMSPeakDetectionMethod.ISOTOPEWAVELET], fm_dict_matching_training=df_dict_matching_training)
    # is always sorted
    prediction_holder = []
    test_class_labels = np.unique(list(test_class_label_dict.values()))
    for pdm, prediction_index in prediction.items():
        predicted_labels = {}
        for p, test_name in zip(prediction_index, test_class_label_dict.keys()):
            predicted_labels[test_name] = test_class_labels[p]
        correct = dict()
        false = dict()
        for fn, predicted_label in predicted_labels.items():
            if predicted_label == test_class_label_dict[fn]:
                correct[fn] = predicted_label
            else:
                false[fn] = predicted_label

        print("resulting_labels for {} are: {}".format(pdm.name, predicted_labels))
        print("Falsely classified: {}".format(false))
        print("That's {} correct vs {} false".format(len(correct.keys()), len(false.keys())))
        prediction_holder.append((predicted_labels, correct, false))
    return prediction_holder


# USER inputs whether centroided or not
if __name__ == "__main__":
    params = {}
    sample_dir = Path("/home/philipp/Desktop/gcms_data/sample")

    sample_path_to_mzml = Path(sample_dir/"FeatureFinderCentroided_1_input.mzML")
    params['input_filename'] = sample_path_to_mzml

    # centroid_sample_path_to_mzml = Path(str(urine_dir) + "/" + "FeatureFinderCentroided_1_input.mzML")
    # params['input_filename'] = centroid_sample_path_to_mzml
    # use same tiny file several times
    # feature_xml_lis = []
    # for i in range(3):
    #     current_sample_path = Path(f"{sample_dir}/{sample_path_to_mzml.stem}_{i}.mzML")
    #     params['input_filename'] = current_sample_path
    #     params = run_centroid_sample(params)
    #     feature_xml_fn = params['feature_storage']
    #     feature_xml_lis.append(feature_xml_fn)
    # feature_df = run_alignment_test(feature_xml_lis)

    # what if there are several experiments in single mzml file? they get huge - several GB - need to rethink indexing by source file
    #   sourceFile / sourceFiles attribute in mzML
    # according to https://github.com/OpenMS/OpenMS/issues/1655 abuse of file format - one run is one mzML file -> not supported

    # save feature matrix
    run_tiny_samples_parallel(3)
    run_algae_platform()


import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="breathpy",
    version="0.8.1",
    description="Breath analysis in python",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/philmaweb/breathpy",
    author="Philipp Weber",
    author_email="pweber@imada.sdu.dk",
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    python_requires='>=3.6',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "graphviz>=0.13.2",
        "ipdb",
        "matplotlib>=3.2.1",
        "matplotlib-venn>=0.11.5",
        "numpy>=1.18.1",
        "pandas>=1.0.3",
        "psutil>=3.4.2",
        "pyopenms==2.4.0",
        "pywavelets>=1.1.1",
        "scikit-image>=0.16.2",
        "scikit-learn>=0.22.0,<0.23.0",
        "scipy>=1.4.1",
        "seaborn>=0.10.0",
        "statannot>=0.2.3",
        "statsmodels>=0.11.1",
        "xlrd>=1.2.0",
    ],
)
